###############################################################################
# Examination of SAR processing software
# Copyright (c) 2019-2021, the pyroSAR Developers.

# This file is part of the pyroSAR Project. It is subject to the
# license terms in the LICENSE.txt file found in the top-level
# directory of this distribution and at
# https://github.com/johntruckenbrodt/pyroSAR/blob/master/LICENSE.txt.
# No part of the pyroSAR project, including this file, may be
# copied, modified, propagated, or distributed except according
# to the terms contained in the LICENSE.txt file.
###############################################################################
import ast
import json
import os
import shutil
import platform
import re
import warnings
import subprocess as sp
import pkg_resources

from pyroSAR._dev_config import ConfigHandler
from spatialist.ancillary import finder

import logging

log = logging.getLogger(__name__)

config = ConfigHandler()


class ExamineSnap(object):
    """
    Class to check if ESA SNAP is installed.
    Upon initialization, this class searches for relevant binaries and the accompanying
    relative directory structure, which uniquely identify an ESA SNAP installation on a system.
    First, all relevant file and folder names are read from the pyroSAR config file if it exists
    and their existence is verified.
    If this fails, a system check is performed to find relevant binaries in the system PATH variable and
    additional files and folders relative to them.
    Furthermore, a snap.auxdata.properties file is scanned for auxiliary data URLs and local storage location.
    This is used by SNAP to manage data from e.g. the SRTM mission. In case SNAP is not installed, the respective
    information is read from a default file delivered with pyroSAR. This has the advantage of using the SNAP download
    URLs and local directory structure without having SNAP installed such that it can be adapted by other SAR software.
    """
    
    def __init__(self):
        
        # define some attributes which identify SNAP
        self.identifiers = ['path', 'gpt', 'etc', 'auxdata']
        
        # a list of relevant sections
        self.sections = ['SNAP', 'OUTPUT', 'SNAP_SUFFIX']
        
        # try reading all necessary attributes from the config file
        self.__read_config()
        
        # if SNAP could not be identified from the config attributes, do a system search for it
        if not self.__is_identified():
            log.debug('identifying SNAP')
            self.__identify_snap()
        
        # if the auxdatapath attribute was not yet set, create a default directory
        if not hasattr(self, 'auxdatapath'):
            self.auxdatapath = os.path.join(os.path.expanduser('~'), '.snap', 'auxdata')
            os.makedirs(self.auxdatapath, exist_ok=True)
        
        # if the SNAP auxdata properties attribute was not yet identified,
        # point it to the default file delivered with pyroSAR
        if not hasattr(self, 'properties'):
            # log.info('using default properties file..')
            template = os.path.join('snap', 'data', 'snap.auxdata.properties')
            self.properties = pkg_resources.resource_filename(__name__, template)
        
        # if the SNAP suffices attribute was not yet identified,
        # point it to the default file delivered with pyroSAR
        if not hasattr(self, 'suffices'):
            template = os.path.join('snap', 'data', 'snap.suffices.properties')
            fname_suffices = pkg_resources.resource_filename(__name__, template)
            with open(fname_suffices, 'r') as infile:
                content = infile.read().split('\n')
            self.__suffices = {k: v for k, v in [x.split('=') for x in content]}
        
        # update the snap properties; this reads the 'properties' file and looks for any changes,
        # which are then updated for the object
        self.__update_snap_properties()
        
        # update the config file: this scans for config changes and re-writes the config file if any are found
        self.__update_config()
    
    def __getattr__(self, item):
        raise AttributeError("'ExamineSnap' object has no attribute '{}'".format(item))
    
    def __is_identified(self):
        """
        Check if SNAP has been properly identified, i.e. all paths in self.identifiers
        have been detected and confirmed.
        
        Returns
        -------
        bool
        """
        return sum([hasattr(self, x) for x in self.identifiers]) == len(self.identifiers)
    
    def __identify_snap(self):
        """
        do a comprehensive search for an ESA SNAP installation
        
        Returns
        -------
        bool
            has the SNAP properties file been changed?
        """
        # create a list of possible SNAP executables
        defaults = ['snap64.exe', 'snap32.exe', 'snap.exe', 'snap']
        paths = os.environ['PATH'].split(os.path.pathsep)
        options = [os.path.join(path, option) for path in paths for option in defaults]
        options = [x for x in options if os.path.isfile(x)]
        
        if not hasattr(self, 'path') or not os.path.isfile(self.path):
            executables = options
        else:
            executables = [self.path] + options
        
        if len(executables) == 0:
            log.debug("could not detect any potential 'snap' executables")
        
        # for each possible SNAP executable, check whether additional files and directories exist relative to it
        # to confirm whether it actually is a ESA SNAP installation or something else like e.g. the Ubuntu App Manager
        for path in executables:
            log.debug('checking candidate {}'.format(path))
            if os.path.islink(path):
                path = os.path.realpath(path)
            
            # check whether a directory etc exists relative to the SNAP executable
            etc = os.path.join(os.path.dirname(os.path.dirname(path)), 'etc')
            if not os.path.isdir(etc):
                log.debug("could not find the 'etc' directory")
                continue
            
            # check the content of the etc directory
            auxdata = os.listdir(etc)
            if 'snap.auxdata.properties' not in auxdata:
                log.debug("could not find the 'snap.auxdata.properties' file")
                continue
            else:
                auxdata_properties = os.path.join(etc, 'snap.auxdata.properties')
            
            # identify the gpt executable
            gpt_candidates = finder(os.path.dirname(path), ['gpt', 'gpt.exe'])
            if len(gpt_candidates) == 0:
                log.debug("could not find the 'gpt' executable")
                continue
            else:
                gpt = gpt_candidates[0]
            
            self.path = path
            self.etc = etc
            self.gpt = gpt
            self.auxdata = auxdata
            self.properties = auxdata_properties
            return
        
        log.warning('SNAP could not be identified. If you have installed it please add the path to the SNAP '
                    'executables (bin subdirectory) to the PATH environment. '
                    'E.g. in the Linux .bashrc file add the following line:\nexport PATH=$PATH:path/to/snap/bin"')
    
    def __read_config(self):
        """
        This method reads the config.ini to examine the snap paths.
        If the snap paths are not in the config.ini or the paths are wrong they will be automatically created.

        Returns
        -------

        """
        for attr in self.identifiers + ['auxdatapath', 'properties']:
            self.__read_config_attr(attr, 'SNAP')
        
        snap_properties = {}
        if 'OUTPUT' in config.sections:
            snap_properties = config['OUTPUT']
        if len(snap_properties.keys()) > 0:
            self.snap_properties = snap_properties
        
        suffices = {}
        if 'SNAP_SUFFIX' in config.sections:
            suffices = config['SNAP_SUFFIX']
        if len(suffices.keys()) > 0:
            self.__suffices = suffices
    
    def __read_config_attr(self, attr, section):
        """
        read an attribute from the config file and set it as an object attribute
        
        Parameters
        ----------
        attr: str
            the attribute name
        section: str
            the config section to read the attribute from
        
        Returns
        -------
        
        """
        if section in config.sections:
            if attr in config[section].keys():
                val = config[section][attr]
                if attr in ['path', 'gpt', 'properties']:
                    exist = os.path.isfile(val)
                elif attr == 'auxdata':
                    val = ast.literal_eval(val)
                    exist = isinstance(val, list)
                else:
                    exist = os.path.isdir(val)
                if exist:
                    # log.info('setting attribute {}'.format(attr))
                    setattr(self, attr, val)
    
    def __update_config(self):
        for section in self.sections:
            if section not in config.sections:
                # log.info('creating section {}..'.format(section))
                config.add_section(section)
        
        for key in self.identifiers + ['auxdatapath', 'properties']:
            if hasattr(self, key):
                self.__update_config_attr(key, getattr(self, key), 'SNAP')
        
        for key, value in self.snap_properties.items():
            self.__update_config_attr(key, value, 'OUTPUT')
        
        for key in sorted(self.__suffices.keys()):
            self.__update_config_attr(key, self.__suffices[key], 'SNAP_SUFFIX')
    
    @staticmethod
    def __update_config_attr(attr, value, section):
        if isinstance(value, list):
            value = json.dumps(value)
        
        if attr not in config[section].keys() or config[section][attr] != value:
            # log.info('updating attribute {0}:{1}..'.format(section, attr))
            # log.info('  {0} -> {1}'.format(repr(config[section][attr]), repr(value)))
            config.set(section, key=attr, value=value, overwrite=True)
    
    def __update_snap_properties(self):
        """
        Read the snap.auxdata.properties file entries to object attributes

        Returns
        -------

        """
        pattern = r'^(?P<key>[\w\.]*)\s*=\s*(?P<value>.*)\n'
        
        if not hasattr(self, 'snap_properties'):
            self.snap_properties = {}
        
        demPath = os.path.join(self.auxdatapath, 'dem')
        landCoverPath = os.path.join(self.auxdatapath, 'LandCover')
        
        with open(self.properties, 'r') as prop:
            for line in prop:
                if re.search(pattern, line):
                    key, value = re.match(re.compile(pattern), line).groups()
                    value = value \
                        .replace('${AuxDataPath}', self.auxdatapath) \
                        .replace('${demPath}', demPath) \
                        .replace('${landCoverPath}', landCoverPath) \
                        .replace('\\', '/')
                    if key not in self.snap_properties.keys() or self.snap_properties[key] != value:
                        self.snap_properties[key] = value
    
    def get_suffix(self, operator):
        """
        get the file name suffix for an operator
        
        Parameters
        ----------
        operator: str
            the name of the operator

        Returns
        -------
        str
            the file suffix
        
        Examples
        --------
        >>> from pyroSAR.examine import ExamineSnap
        >>> config = ExamineSnap()
        >>> print(config.get_suffix('Terrain-Flattening'))
        'TF'
        """
        if operator in self.__suffices.keys():
            return self.__suffices[operator]
        else:
            return None
    
    def get_version(self, module):
        """
        Read the version and date of different SNAP modules.
        This scans a file 'messages.log', which is re-written every time SNAP is started.
        
        Parameters
        ----------
        module: str
            one of the following
            
            - core
            - desktop
            - rstbx
            - s1tbx
            - s2tbx
            - s3tbx

        Returns
        -------
        dict
            a dictionary with keys 'version' and 'date'
        """
        # base search patterns for finding the right lines
        patterns = {'core': r'org\.esa\.snap\.snap\.core',
                    'desktop': r'org\.esa\.snap\.snap\.ui',
                    'rstb': r'org\.csa\.rstb\.rstb\.kit',
                    's1tbx': r'org\.esa\.s1tbx\.s1tbx\.kit',
                    's2tbx': r'org\.esa\.s2tbx\.s2tbx\.kit',
                    's3tbx': r'org\.esa\.s3tbx\.s3tbx\.kit'}
        
        if module in patterns.keys():
            pattern = patterns[module]
            pattern += r' \[(?P<version>[0-9.]+) [0-9.]+ (?P<date>[0-9]{12})'
        else:
            raise RuntimeError('module not supported')
        
        system = platform.system()
        if system in ['Linux', 'Darwin']:
            path = os.path.join(os.path.expanduser('~'), '.snap', 'system')
        elif system == 'Windows':
            path = os.path.join(os.environ['APPDATA'], 'SNAP')
        else:
            raise RuntimeError('operating system not supported')
        
        conda_env_path = os.environ.get('CONDA_PREFIX')
        if conda_env_path is not None and conda_env_path in self.gpt:
            fname = os.path.join(conda_env_path, 'snap', '.snap', 'system', 'var', 'log', 'messages.log')
        else:
            fname = os.path.join(path, 'var', 'log', 'messages.log')
        
        if not os.path.isfile(fname):
            try:
                # This will start SNAP and immediately stop it because of the invalid argument.
                # Currently this seems to be the only way to create the messages.log file if it does not exist.
                sp.check_call([self.path, '--nosplash', '--dummytest', '--console', 'suppress'])
            except sp.CalledProcessError:
                pass
        
        if not os.path.isfile(fname):
            raise RuntimeError("cannot find 'messages.log' to read SNAP module versions from.")
        
        with open(fname, 'r') as m:
            content = m.read()
        match = re.search(pattern, content)
        if match is None:
            raise RuntimeError('cannot read version information from {}.\nPlease restart SNAP.'.format(fname))
        return match.groupdict()


class ExamineGamma(object):
    def __init__(self):
        home_sys = os.environ.get('GAMMA_HOME')
        if home_sys is not None and not os.path.isdir(home_sys):
            warnings.warn('found GAMMA_HOME environment variable, but directory does not exist')
            home_sys = None
        
        self.__read_config()
        
        if hasattr(self, 'home'):
            if home_sys is not None and self.home != home_sys:
                log.info('the value of GAMMA_HOME is different to that in the pyroSAR configuration;\n'
                         '  was: {}\n'
                         '  is : {}\n'
                         'resetting the configuration and deleting parsed modules'
                         .format(self.home, home_sys))
                parsed = os.path.join(os.path.dirname(self.fname), 'gammaparse')
                shutil.rmtree(parsed)
                self.home = home_sys
        if not hasattr(self, 'home'):
            if home_sys is not None:
                setattr(self, 'home', home_sys)
            else:
                raise RuntimeError('could not read GAMMA installation directory')
        self.version = re.search('GAMMA_SOFTWARE-(?P<version>[0-9]{8})',
                                 getattr(self, 'home')).group('version')
        
        if not hasattr(self, 'gdal_config'):
            gdal_config = '/usr/bin/gdal-config'
            if not os.path.isfile(gdal_config):
                warnings.warn('could not find GDAL system installation under {}\n'
                              'please modify the pyroSAR configuration: {}'.format(gdal_config, self.fname))
            self.gdal_config = gdal_config
        self.__update_config()
    
    def __read_config(self):
        self.fname = config.file
        if 'GAMMA' in config.sections:
            attr = config['GAMMA']
            for key, value in attr.items():
                setattr(self, key, value)
    
    def __update_config(self):
        if 'GAMMA' not in config.sections:
            config.add_section('GAMMA')
        
        for attr in ['home', 'version', 'gdal_config']:
            self.__update_config_attr(attr, getattr(self, attr), 'GAMMA')
    
    @staticmethod
    def __update_config_attr(attr, value, section):
        if isinstance(value, list):
            value = json.dumps(value)
        
        if attr not in config[section].keys() or config[section][attr] != value:
            config.set(section, key=attr, value=value, overwrite=True)



###############################################################################
# ancillary routines for software pyroSAR

# Copyright (c) 2014-2022, the pyroSAR Developers.

# This file is part of the pyroSAR Project. It is subject to the
# license terms in the LICENSE.txt file found in the top-level
# directory of this distribution and at
# https://github.com/johntruckenbrodt/pyroSAR/blob/master/LICENSE.txt.
# No part of the pyroSAR project, including this file, may be
# copied, modified, propagated, or distributed except according
# to the terms contained in the LICENSE.txt file.
###############################################################################
"""
This module gathers central functions and classes for general pyroSAR applications.
"""
import os
import re
from math import sin, radians
import inspect
from datetime import datetime
from ._dev_config import product_pattern

from spatialist.ancillary import finder


def groupby(images, attribute):
    """
    group a list of images by a metadata attribute
    
    Parameters
    ----------
    images: list[str]
        the names of the images to be sorted
    attribute: str
        the name of the attribute used for sorting;
        see :func:`parse_datasetname` for options
    
    Returns
    -------
    list[list[str]]
        a list of sub-lists containing the grouped images
    """
    images_sort = sorted(images, key=lambda x: re.search(product_pattern, x).group(attribute))
    out_meta = [[parse_datasetname(images_sort.pop(0))]]
    while len(images_sort) > 0:
        filename = images_sort.pop(0)
        meta = parse_datasetname(filename)
        
        if out_meta[-1][0][attribute] == meta[attribute]:
            out_meta[-1].append(meta)
        else:
            out_meta.append([meta])
    out = [[x['filename'] for x in y] for y in out_meta]
    return out


def groupbyTime(images, function, time):
    """
    function to group images by their acquisition time difference

    Parameters
    ----------
    images: list[str]
        a list of image names
    function: function
        a function to derive the time from the image names; see e.g. :func:`seconds`
    time: int or float
        a time difference in seconds by which to group the images

    Returns
    -------
    list[list[str]]
        a list of sub-lists containing the grouped images
    """
    # sort images by time stamp
    srcfiles = sorted(images, key=function)
    
    groups = [[srcfiles[0]]]
    group = groups[0]
    
    for i in range(1, len(srcfiles)):
        item = srcfiles[i]
        timediff = abs(function(item) - function(group[-1]))
        if timediff <= time:
            group.append(item)
        else:
            groups.append([item])
            group = groups[-1]
    return [x[0] if len(x) == 1 else x for x in groups]


def multilook_factors(source_rg, source_az, target, geometry, incidence):
    """
    compute multi-looking factors to approximate a square pixel with defined target ground range pixel spacing.
    
    Parameters
    ----------
    source_rg: int or float
        the range pixel spacing
    source_az: int or float
        the azimuth pixel spacing
    target: int or float
        the target pixel spacing of an approximately square pixel
    geometry: str
        the imaging geometry; either 'SLANT_RANGE' or 'GROUND_RANGE'
    incidence: int or float
        the angle of incidence

    Returns
    -------
    tuple[int]
        the multi-looking factors as (range looks, azimuth looks)
    
    Examples
    --------
    >>> from pyroSAR.ancillary import multilook_factors
    >>> rlks, azlks = multilook_factors(source_rg=2, source_az=13, target=10,
    >>>                                 geometry='SLANT_RANGE', incidence=39)
    >>> print(rlks, azlks)
    4 1
    """
    azlks = int(round(float(target) / source_az))
    azlks = azlks if azlks > 0 else 1
    if geometry == 'SLANT_RANGE':
        rlks = float(azlks) * source_az * sin(radians(incidence)) / source_rg
    elif geometry == 'GROUND_RANGE':
        rlks = float(azlks) * source_az / source_rg
    else:
        raise ValueError("parameter 'geometry' must be either 'SLANT_RANGE' or 'GROUND_RANGE'")
    
    rlks = int(round(rlks))
    return rlks, azlks


def seconds(filename):
    """
    function to extract time in seconds from a file name.
    the format must follow a fixed pattern: YYYYmmddTHHMMSS
    Images processed with pyroSAR functionalities via module snap or gamma will contain this information.

    Parameters
    ----------
    filename: str
        the name of a file from which to extract the time from

    Returns
    -------
    float
        the difference between the time stamp in filename and Jan 01 1900 in seconds
    """
    # return mktime(strptime(re.findall('[0-9T]{15}', filename)[0], '%Y%m%dT%H%M%S'))
    td = datetime.strptime(re.findall('[0-9T]{15}', filename)[0], '%Y%m%dT%H%M%S') - datetime(1900, 1, 1)
    return td.total_seconds()


def parse_datasetname(name, parse_date=False):
    """
    Parse the name of a pyroSAR processing product and extract its metadata components as dictionary
    
    Parameters
    ----------
    name: str
        the name of the file to be parsed
    parse_date: bool
        parse the start date to a :class:`~datetime.datetime` object or just return the string?
    
    Returns
    -------
    dict
        the metadata attributes
    
    Examples
    --------
    >>> meta = parse_datasetname('S1A__IW___A_20150309T173017_VV_grd_mli_geo_norm_db.tif')
    >>> print(sorted(meta.keys()))
    ['acquisition_mode', 'extensions', 'filename', 'orbit',
    'outname_base', 'polarization', 'proc_steps', 'sensor', 'start']
    """
    
    filename = os.path.abspath(name) if os.path.isfile(name) else name
    
    match = re.match(re.compile(product_pattern), filename)
    if not match:
        return
    out = match.groupdict()
    if out['extensions'] == '':
        out['extensions'] = None
    if out['proc_steps'] is not None:
        out['proc_steps'] = out['proc_steps'].split('_')
    if parse_date:
        out['start'] = datetime.strptime(out['start'], '%Y%m%dT%H%M%S')
    out['filename'] = filename
    out['outname_base'] = out['outname_base'].strip('_')
    return out


def find_datasets(directory, recursive=False, **kwargs):
    """
    find pyroSAR datasets in a directory based on their metadata
    
    Parameters
    ----------
    directory: str
        the name of the directory to be searched
    recursive: bool
        search the directory recursively into subdirectories?
    kwargs:
        Metadata attributes for filtering the scene list supplied as `key=value`. e.g. `sensor='S1A'`.
        Multiple allowed options can be provided in tuples, e.g. `sensor=('S1A', 'S1B')`.
        Any types other than tuples require an exact match, e.g. `proc_steps=['grd', 'mli', 'geo', 'norm', 'db']`
        will be matched only if these processing steps are contained in the product name in this exact order.
        The special attributes `start` and `stop` can be used for time filtering where `start<=value<=stop`.
        See function :func:`parse_datasetname` for further options.
    
    Returns
    -------
    list of str
        the file names found in the directory and filtered by metadata attributes
    
    Examples
    --------
    >>> selection = find_datasets('path/to/files', sensor=('S1A', 'S1B'), polarization='VV')
    """
    files = finder(directory, [product_pattern], regex=True, recursive=recursive)
    selection = []
    for file in files:
        meta = parse_datasetname(file)
        matches = []
        for key, val in kwargs.items():
            if key == 'start':
                match = val <= meta['start']
            elif key == 'stop':
                match = val >= meta['start']  # only the start time stamp is contained in the filename
            elif isinstance(val, tuple):
                match = meta[key] in val
            else:
                match = meta[key] == val
            matches.append(match)
        if all(matches):
            selection.append(file)
    return selection


def getargs(func):
    """
    get the arguments of a function
    
    Parameters
    ----------
    func: function
        the function to be checked

    Returns
    -------
    list or str
        the argument names
    """
    return sorted(inspect.getfullargspec(func).args)


def hasarg(func, arg):
    """
    simple check whether a function takes a parameter as input
    
    Parameters
    ----------
    func: function
        the function to be checked
    arg: str
        the argument name to be found

    Returns
    -------
    bool
        does the function take this as argument?
    """
    return arg in getargs(func)


def windows_fileprefix(func, path, exc_info):
    """
    Helper function for :func:`shutil.rmtree` to exceed Windows' file name length limit of 256 characters.
    See `here <https://stackoverflow.com/questions/36219317/pathname-too-long-to-open>`_ for details.

    Parameters
    ----------
    func: function
        the function to be executed, i.e. :func:`shutil.rmtree`
    path: str
        the path to be deleted
    exc_info: tuple
        execution info as returned by :func:`sys.exc_info`

    Returns
    -------

    Examples
    --------
    >>> import shutil
    >>> from pyroSAR.ancillary import windows_fileprefix
    >>> shutil.rmtree('/path', onerror=windows_fileprefix)
    """
    func(u'\\\\?\\' + path)



###############################################################################
# interface for translating GAMMA errors messages into Python error types

# Copyright (c) 2015-2019, the pyroSAR Developers.

# This file is part of the pyroSAR Project. It is subject to the
# license terms in the LICENSE.txt file found in the top-level
# directory of this distribution and at
# https://github.com/johntruckenbrodt/pyroSAR/blob/master/LICENSE.txt.
# No part of the pyroSAR project, including this file, may be
# copied, modified, propagated, or distributed except according
# to the terms contained in the LICENSE.txt file.
###############################################################################

import re


def gammaErrorHandler(out, err):
    """
    Function to raise errors in Python. This function is not intended for direct use, but as part of function gamma.util.process
    Args:
        out: the stdout message returned by a subprocess call of a gamma command
        err: the stderr message returned by a subprocess call of a gamma command

    Raises: IOError | ValueError | RuntimeError | None

    """
    
    # scan stdout and stdin messages for lines starting with 'ERROR'
    messages = out.split('\n') if out else []
    messages.extend(err.strip().split('\n'))
    errormessages = [x for x in messages if x.startswith('ERROR')]
    
    # registry of known gamma error messages and corresponding Python error types
    # do not change the Python error types of specific messages! This will change the behavior of several functions
    # in case no error is to be thrown define None as error type
    knownErrors = {'image data formats differ': IOError,
                   'cannot open': IOError,
                   r'no coverage of SAR image by DEM(?: \(in (?:latitude/northing|longitude/easting)\)|)': RuntimeError,
                   'libgdal.so.1: no version information available': None,
                   'line outside of image': ValueError,
                   'no offsets found above SNR threshold': ValueError,
                   'window size < 4': ValueError,
                   'MLI oversampling factor must be 1, 2, 4, 8': ValueError,
                   'no points available for determining average intensity': ValueError,
                   'p_interp(): time outside of range': RuntimeError,
                   'no overlap with lookup table': RuntimeError,
                   'insufficient offset points to determine offset model parameters': RuntimeError,
                   'insufficient offset points left after culling to determine offset model parameters': RuntimeError,
                   'calloc_1d: number of elements <= 0': ValueError,
                   'multi-look output line:': RuntimeError,
                   'no OPOD state vector found with the required start time!': RuntimeError,
                   'gc_map operates only with slant range geometry, image geometry in SLC_par: GROUND_RANGE': RuntimeError,
                   'OPOD state vector data ends before start of the state vector time window': RuntimeError,
                   'non-zero exit status': RuntimeError,
                   'unsupported DEM projection': RuntimeError,
                   'tiffWriteProc:No space left on device': RuntimeError,
                   'in subroutine julday: there is no year zero!': RuntimeError,
                   'cannot create ISP image parameter file': OSError}
    
    # check if the error message is known and throw the mapped error from knownErrors accordingly.
    # Otherwise throw an GammaUnknownError.
    # The actual message is passed to the error and thus visible for backtracing
    if len(errormessages) > 0:
        errormessage = errormessages[-1]
        err_out = '\n\n'.join([re.sub('ERROR[: ]*', '', x) for x in errormessages])
        for error in knownErrors:
            if re.search(error, errormessage):
                errortype = knownErrors[error]
                if errortype:
                    raise errortype(err_out)
                else:
                    return
        raise GammaUnknownError(err_out)


class GammaUnknownError(Exception):
    """
    This is a general error, which is raised if the error message is not yet integrated
    into the known errors of function gammaErrorHandler.
    If this error occurs the message should be included in function gammaErrorHandler.
    """
    
    def __init__(self, errormessage):
        Exception.__init__(self, errormessage)



###############################################################################
# Convenience tools for Open Data Cube ingestion

# Copyright (c) 2018-2019, the pyroSAR Developers.

# This file is part of the pyroSAR Project. It is subject to the
# license terms in the LICENSE.txt file found in the top-level
# directory of this distribution and at
# https://github.com/johntruckenbrodt/pyroSAR/blob/master/LICENSE.txt.
# No part of the pyroSAR project, including this file, may be
# copied, modified, propagated, or distributed except according
# to the terms contained in the LICENSE.txt file.
###############################################################################
"""
This (still experimental) module is intended to easily prepare SAR scenes processed
by pyroSAR for ingestion into an Open Data Cube.

.. code-block:: python

    from pyroSAR.datacube_util import Product, Dataset
    from pyroSAR.ancillary import find_datasets
    
    # find pyroSAR files by metadata attributes
    archive_s1 = '/.../sentinel1/GRD/processed'
    scenes_s1 = find_datasets(archive_s1, sensor=('S1A', 'S1B'), acquisition_mode='IW')
    
    # group the found files by their file basenames
    # files with the same basename are considered to belong to the same dataset
    grouped = groupby(scenes_s1, 'outname_base')
    
    # define the polarization units describing the data sets
    units = {'VV': 'backscatter VV', 'VH': 'backscatter VH'}
    
    # create a new product
    with Product(name='S1_GRD_index',
                 product_type='gamma0',
                 description='Gamma Naught RTC backscatter') as prod:
        
        for dataset in grouped:
            with Dataset(dataset, units=units) as ds:
                
                # add the dataset to the product
                prod.add(ds)
                
                # parse datacube indexing YMLs from product and data set metadata
                prod.export_indexing_yml(ds, 'yml_index_outdir')
        
        # write the product YML
        prod.write('yml_product')
        
        # print the product metadata which is written to the product YML
        print(prod)
"""

import os
import re
import yaml
import uuid
from time import strftime, strptime
from spatialist.raster import Raster, Dtype
from spatialist.ancillary import union
from .ancillary import parse_datasetname

import logging
log = logging.getLogger(__name__)


class Dataset(object):
    """
    A general class describing dataset information required for creating ODC YML files
    
    Parameters
    ----------
    filename: str, list, Dataset
        the product to be used; either an existing :class:`Dataset` object or a (list of) file(s) matching the pyroSAR
        naming pattern, i.e. that can be parsed by :func:`pyroSAR.ancillary.parse_datasetname`
    units: str or dict
        the units of the product measurement
    """
    def __init__(self, filename, units='DN'):
        
        if isinstance(filename, list):
            combined = sum([Dataset(x, units) for x in filename])
            self.__init__(combined)
        
        elif isinstance(filename, Dataset):
            for attr, value in vars(filename).items():
                setattr(self, attr, value)
        
        elif isinstance(filename, str):
            # map pyroSAR sensor identifiers to platform and instrument codes
            sensor_lookup = {'ASAR': ('ENVISAT', 'ASAR'),
                             'ERS1': ('ERS-1', 'SAR'),
                             'ERS2': ('ERS-2', 'SAR'),
                             'PSR1': ('ALOS-1', 'PALSAR'),
                             'PSR2': ('ALOS-2', 'PALSAR-2'),
                             'S1A': ('SENTINEL-1', 'C-SAR'),
                             'S1B': ('SENTINEL-1', 'C-SAR'),
                             'TSX1': ('TERRASAR-X_1', 'SAR'),
                             'TDX1': ('TANDEM-X_1', 'SAR')}
            
            # extract basic metadata attributes from the filename and register them to the object
            meta = parse_datasetname(filename)
            
            if meta is None:
                raise ValueError('could not identify dataset: {}'.format(filename))
            
            for key, val in meta.items():
                setattr(self, key, val)
            
            # define acquisition start and end time; Currently both are set to the acquisition start time,
            # which is contained in the filename
            # Time will only be correct if the full scene was processed, start and end time of s subset will
            # differ. Thus, accurately setting both is not seen as too relevant.
            self.from_dt = strftime('%Y-%m-%dT%H:%M:%S', strptime(self.start, '%Y%m%dT%H%M%S'))
            self.to_dt = strftime('%Y-%m-%dT%H:%M:%S', strptime(self.start, '%Y%m%dT%H%M%S'))
            
            # match the sensor ID from the filename to a platform and instrument
            if self.sensor not in sensor_lookup.keys():
                raise ValueError('unknown sensor: {}'.format(self.sensor))
            
            self.platform, self.instrument = sensor_lookup[self.sensor]
            
            # extract general geo metadata from the GTiff information
            with Raster(filename) as ras:
                self.dtype = Dtype(ras.dtype).numpystr
                self.nodata = ras.nodata
                self.format = ras.format
                self.xres, self.yres = ras.res
                self.crs = 'EPSG:{}'.format(ras.epsg)
                self.is_projected = ras.projcs is not None
                self.extent = self.__extent_convert(ras.geo, 'x', 'y')
                # reproject the raster bounding box to EPSG 4326 and store its extent
                with ras.bbox() as bbox:
                    bbox.reproject(4326)
                    self.extent_4326 = self.__extent_convert(bbox.extent, 'lon', 'lat')
            
            # create dictionary for resolution metadata depending on CRS characteristics
            resolution_keys = ('x', 'y') if self.is_projected else ('longitude', 'latitude')
            self.resolution = dict(zip(resolution_keys, (self.xres, self.yres)))
            
            # check whether the data type is supported
            pattern = '(?:(?:u|)int(?:8|16|32|64)|float(?:32|64))'
            if not re.search(pattern, self.dtype):
                raise ValueError('unsupported data type {}'.format(self.dtype))
            
            # determine the dataset units
            if isinstance(units, str):
                units = units
            elif isinstance(units, dict):
                try:
                    units = units[self.polarization]
                except KeyError:
                    raise KeyError("parameter 'units' does not contain key '{}'".format(self.polarization))
            else:
                raise TypeError("parameter 'units' must be of type str or dict")
            
            # create the measurement entry from collected metadata;
            # this is intended for easy access by class Product
            self.measurements = {self.polarization: {'dtype': self.dtype,
                                                     'name': self.polarization,
                                                     'nodata': self.nodata,
                                                     'filename': filename,
                                                     'units': units}}
        else:
            raise TypeError('filename must be of type str, list or Dataset')
    
    def __add__(self, dataset):
        """
        override the + operator. This is intended to easily combine two Dataset objects, which were
        created from different files belonging to the same measurement, e.g. two GeoTIFFs with one polarization
        each.
        
        Parameters
        ----------
        dataset: Dataset
            the dataset to add to the current one

        Returns
        -------
        Dataset
            the combination of the two
        """
        for attr in ['extent', 'crs', 'sensor', 'acquisition_mode', 'proc_steps', 'outname_base']:
            if getattr(self, attr) != getattr(dataset, attr):
                raise ValueError('value mismatch: {}'.format(attr))
        # self.filename.append(dataset.filename)
        for key in dataset.measurements.keys():
            if key in self.measurements.keys():
                raise RuntimeError('only different measurements can be combined to one dataset')
        self.measurements.update(dataset.measurements)
        return self
    
    def __radd__(self, dataset):
        """
        similar to :meth:`Dataset.__add__` but for function :func:`sum`, e.g. :code:`sum([Dataset1, Dataset2])`
        
        Parameters
        ----------
        dataset: Dataset
            the dataset to add to the current one

        Returns
        -------
        Dataset
            the combination of the two
        """
        if dataset == 0:
            return self
        else:
            return self.__add__(dataset)
    
    @staticmethod
    def __extent_convert(extent, xkey, ykey):
        """
        convert the extent of a :class:`~spatialist.raster.Raster` object to a
        datacube-compliant dictionary.
        
        Parameters
        ----------
        extent: dict
            the extent as returned by a :class:`~spatialist.raster.Raster` object
        xkey: {'longitude', 'x'}
            the key of the x dimension
        ykey: {'latitude', 'y'}
            the key of the y dimension

        Returns
        -------
        dict
            a dictionary with keys `ll`, `lr`, `ul` and ``ur
        """
        return {'ll': {xkey: extent['xmin'],
                       ykey: extent['ymin']},
                'lr': {xkey: extent['xmax'],
                       ykey: extent['ymin']},
                'ul': {xkey: extent['xmin'],
                       ykey: extent['ymax']},
                'ur': {xkey: extent['xmax'],
                       ykey: extent['ymax']}}
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
    
    def __get_measurement_attr(self, attr):
        """
        get a certain measurement attribute from all measurements
        
        Parameters
        ----------
        attr: str
            the attribute to get

        Returns
        -------
        dict
            a dictionary with the measurement names as keys and the respective attribute as value
        """
        return dict([(key, self.measurements[key][attr]) for key in self.measurements.keys()])
    
    @property
    def filenames(self):
        """
        
        Returns
        -------
        dict
            all file names registered in the dataset
        """
        return self.__get_measurement_attr('filename')
    
    @property
    def identifier(self):
        """
        
        Returns
        -------
        str
            a unique dataset identifier
        """
        return '{}_{}'.format(self.outname_base, '_'.join(self.proc_steps))
    
    @property
    def units(self):
        """
        
        Returns
        -------
        dict
            all measurement unit names registered in the dataset
        """
        return self.__get_measurement_attr('units')
    
    @units.setter
    def units(self, value):
        """
        (re)set the units of all measurements
        
        Parameters
        ----------
        value: str or dict
            the unit(s) to be set; if multiple measurements are present,
            a dictionary with measurement names as keys needs to be defined

        Returns
        -------

        """
        keys = list(self.measurements.keys())
        if isinstance(value, str):
            if len(keys) == 1:
                self.measurements[keys[0]]['units'] = value
            else:
                raise TypeError('the dataset contains multiple measurements; '
                                'in this case a dictionary is needed for setting the measurement units')
        elif isinstance(value, dict):
            for name, unit in value.items():
                if name in keys:
                    self.measurements[name]['units'] = unit
                else:
                    raise KeyError("the dataset does not contain a measurement '{}'".format(name))
    
    def close(self):
        return


class Product(object):
    """
    A class for describing an ODC product definition
    
    Parameters
    ----------
    definition: str, list, None
        the source of the product definition; either an existing product YML, a list of :class:`Dataset` objects,
        or None. In the latter case the product is defined using the parameters
        `name`, `product_type` and `description`.
    name: str
        the name of the product in the data cube
    product_type: str
        the type of measurement defined in the product, e.g. `gamma0`
    description: str
        the description of the product and its measurements
    """
    def __init__(self, definition=None, name=None, product_type=None,
                 description=None):
        
        missing_message = "when initializing {}, parameters " \
                          "'name', 'product_type' and 'description' must be defined"
        
        if isinstance(definition, str):
            if os.path.isfile(definition):
                with open(definition, 'r') as yml:
                    try:
                        self.meta = yaml.load(yml, Loader=yaml.FullLoader)
                    except yaml.YAMLError:
                        raise RuntimeError('the provided file does not seem to be a YAML file')
            else:
                raise RuntimeError('definition file does not exist')
        
        elif isinstance(definition, list):
            if None in [name, product_type, description]:
                raise ValueError(missing_message.format(' a product from list'))
            self.__initialize(name, product_type, description)
            for dataset in definition:
                with Dataset(dataset) as DS:
                    self.add(DS)
        
        elif definition is None:
            if None in [name, product_type, description]:
                raise ValueError(missing_message.format('a blank product'))
            self.__initialize(name, product_type, description)
        else:
            raise TypeError('type of parameter definition must be either str, list or None')
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
    
    def __str__(self):
        return yaml.dump(self.meta, default_flow_style=False)
    
    def __getattr__(self, item):
        if item in self.__fixture_storage:
            return self.meta['storage'][item]
        elif item in self.__fixture_metadata:
            subkey = 'code' if item == 'platform' else 'name'
            return self.meta['metadata'][item][subkey]
        elif item == 'product_type':
            return self.meta['metadata']['product_type']
        else:
            return object.__getattribute__(self, item)
    
    def __setattr__(self, key, value):
        if key in self.__fixture_storage:
            self.meta['storage'][key] = value
        elif key in self.__fixture_metadata:
            subkey = 'code' if key == 'platform' else 'name'
            self.meta['metadata'][key][subkey] = value
        elif key == 'product_type':
            self.meta['metadata']['product_type'] = value
        else:
            super(Product, self).__setattr__(key, value)
    
    def close(self):
        return
    
    def __add_measurement(self, name, dtype, nodata, units):
        """
        create a new measurement entry
        
        Parameters
        ----------
        name: str
            the measurement name
        dtype: str
            the data type, e.g. float32
        nodata: int or float
            the nodata value of the data
        units: str
            the measurement units

        Returns
        -------

        """
        if name in self.measurements.keys():
            raise IndexError('measurement {} already exists'.format(name))
        self.meta['measurements'].append({'name': name,
                                          'dtype': dtype,
                                          'units': units,
                                          'nodata': nodata})
    
    def __initialize(self, name, product_type, description):
        """
        create a new blank product
        
        Parameters
        ----------
        name: str
            the name of the product
        product_type: str
            the product type, e.g. `gamma0`
        description: str
            a description of the product content/purpose

        Returns
        -------

        """
        self.meta = {'description': description,
                     'measurements': [],
                     'metadata': {'platform': {'code': None},
                                  'instrument': {'name': None},
                                  'format': {'name': None},
                                  'product_type': product_type},
                     'metadata_type': 'eo',
                     'name': name,
                     'storage': {'crs': None,
                                 'resolution': None}}
    
    @staticmethod
    def __check_dict_keys(keys, reference):
        return len(union(keys, reference)) == len(keys)
    
    @property
    def __fixture_fields(self):
        """
        
        Returns
        -------
        list
            the names of the top-level metadata fields, which must be defined
        """
        return ['description', 'measurements', 'metadata', 'metadata_type', 'name', 'storage']
    
    @property
    def __fixture_measurement(self):
        """
        
        Returns
        -------
        list
            the names of the metadata fields, which must be defined for all measurements
        """
        return ['dtype', 'nodata', 'units']
    
    @property
    def __fixture_metadata(self):
        """
        
        Returns
        -------
        list
            the names of the metadata fields, which must be defined in the general metadata section
        """
        return ['format', 'instrument', 'platform']
    
    @property
    def __fixture_storage(self):
        """
        
        Returns
        -------
        list
            the names of the metadata fields, which must be defined for the storage section
        """
        return ['crs', 'resolution']
    
    def __validate(self):
        """
        assert whether the Product is valid
        
        Returns
        -------
        
        Raises
        ------
        RuntimeError
        """
        try:
            assert isinstance(self.meta, dict)
            assert self.__check_dict_keys(self.__fixture_fields, self.meta.keys())
            assert 'product_type' in self.meta['metadata'].keys()
            for measurement in self.meta['measurements']:
                assert self.__check_dict_keys(self.__fixture_measurement, measurement.keys())
        except AssertionError as e:
            log.info(e)
            raise RuntimeError('product invalid')
    
    def add(self, dataset):
        """
        Add a dataset to the abstracted product description. This first performs a check
        whether the dataset is compatible with the product and its already existing measurements.
        If a measurement in the dataset does not yet exist in the product description it is added.
        
        Parameters
        ----------
        dataset: Dataset
            the dataset whose description is to be added

        Returns
        -------

        """
        if not isinstance(dataset, Dataset):
            raise TypeError('input must be of type pyroSAR.datacube.Dataset')
        self.check_integrity(dataset, allow_new_measurements=True)
        
        # set the general product definition attributes if they are None
        for attr in self.__fixture_metadata + self.__fixture_storage:
            if getattr(self, attr) is None:
                setattr(self, attr, getattr(dataset, attr))
        
        # if it is not yet present, add the dataset measurement definition to that of the product
        for measurement, content in dataset.measurements.items():
            if measurement not in self.measurements.keys():
                self.__add_measurement(dtype=content['dtype'],
                                       name=content['name'],
                                       nodata=content['nodata'],
                                       units=content['units'])
    
    def check_integrity(self, dataset, allow_new_measurements=False):
        """
        check if a dataset is compatible with the product definition.
        
        Parameters
        ----------
        dataset: Dataset
            the dataset to be checked
        allow_new_measurements: bool
            allow new measurements to be added to the product definition?
            If not and the dataset contains measurements,
            which are not defined in the product, an error is raised.

        Returns
        -------
        
        Raises
        ------
        RuntimeError
        """
        # check general metadata and storage fields
        for attr in self.__fixture_metadata + self.__fixture_storage:
            val_ds = getattr(dataset, attr)
            val_prod = getattr(self, attr)
            if val_prod is not None and val_ds != val_prod:
                raise RuntimeError("mismatch of attribute '{0}': {1}, {2}".format(attr, val_ds, val_prod))
        
        # check measurement fields
        for measurement, content in dataset.measurements.items():
            if measurement not in self.measurements.keys():
                if not allow_new_measurements:
                    raise RuntimeError("measurement '{}' is not present in the product definition "
                                       "and allow_new_measurements is set to False".format(measurement))
            else:
                match = self.measurements[measurement]
                for attr in self.__fixture_measurement:
                    if match[attr] != content[attr]:
                        raise RuntimeError("mismatch of measurement '{0}', "
                                           "attribute '{1}': {2}, {3}".
                                           format(measurement, attr, match[attr], content[attr]))
    
    def export_indexing_yml(self, dataset, outdir):
        """
        Write a YML file named {:meth:`Dataset.identifier`}_dcindex.yml, which can be used for indexing a dataset in
        an Open Data Cube. The file will contain information from the product and the dataset and a test is first
        performed to check whether the dataset matches the product definition.
        A unique ID is issued using :func:`uuid.uuid4()`.
        
        Parameters
        ----------
        dataset: Dataset
            the dataset for which to export a file for
        outdir: str
            the directory to write the file to

        Returns
        -------

        """
        
        self.__validate()
        
        outname = os.path.join(outdir, dataset.identifier + '_dcindex.yml')
        
        if os.path.isfile(outname):
            raise RuntimeError('indexing YML already exists: \n   {}'.format(outname))
        
        if not os.path.isdir(outdir):
            os.makedirs(outdir)
        
        self.check_integrity(dataset)
        out = {'id': str(uuid.uuid4()),
               'image': {'bands': {}},
               'grid_spatial': {'projection': {}},
               'extent': {'coord': {}},
               'lineage': {'source_datasets': {}}}
        
        for measurement, content in dataset.measurements.items():
            out['image']['bands'][measurement] = {'path': content['filename']}
        
        for attr in self.__fixture_metadata:
            subkey = 'code' if attr == 'platform' else 'name'
            out[attr] = {subkey: getattr(dataset, attr)}
        
        out['grid_spatial']['projection']['geo_ref_points'] = dataset.extent
        out['grid_spatial']['projection']['spatial_reference'] = dataset.crs
        
        out['extent']['coord'] = dataset.extent_4326
        out['extent']['from_dt'] = dataset.from_dt
        out['extent']['to_dt'] = dataset.to_dt
        
        out['product_type'] = self.meta['metadata']['product_type']
        
        with open(outname, 'w') as yml:
            yaml.dump(out, yml, default_flow_style=False)
    
    def export_ingestion_yml(self, outname, product_name, ingest_location, chunking):
        """
        Write a YML file, which can be used for ingesting indexed datasets into an Open Data Cube.
        
        Parameters
        ----------
        outname: str
            the name of the YML file to write
        product_name: str
            the name of the product in the ODC
        ingest_location: str
            the location of the ingested NetCDF files
        chunking: dict
            a dictionary with keys 'x', 'y' and 'time'; determines the size of the netCDF
            files ingested into the datacube; e.g. {'x': 512, 'y': 512, 'time': 1}

        Returns
        -------

        """
        if os.path.isfile(outname):
            raise RuntimeError('product definition YML already exists: \n   {}'.format(outname))
        
        self.__validate()
        
        if product_name == self.meta['name']:
            raise ValueError('source and target product names must be different')

        outdir = os.path.dirname(outname)
        if not os.path.isdir(outdir):
            os.makedirs(outdir)
        
        file_path_template = '{0}/{1}_{2}_{3}_{4}_' \
                             '{{tile_index[0]}}_' \
                             '{{tile_index[1]}}_' \
                             '{{start_time}}.nc'.format(product_name,
                                                        self.platform,
                                                        self.instrument,
                                                        self.product_type,
                                                        self.crs.replace('EPSG:', ''))
        
        global_attributes = {'instrument': self.instrument,
                             'platform': self.platform,
                             'institution': 'ESA',
                             'achknowledgment': 'Sentinel-1 data is provided by the European Space Agency '
                                                'on behalf of the European Commission via download.'}
        
        storage = self.meta['storage']
        storage['driver'] = 'NetCDF CF'
        storage['tile_size'] = {}
        storage['tile_size']['x'] = storage['resolution']['x'] * chunking['x']
        storage['tile_size']['y'] = storage['resolution']['y'] * chunking['y']
        storage['chunking'] = chunking
        storage['dimension_order'] = ['time', 'y', 'x']
        
        measurements = self.meta['measurements']
        for measurement in measurements:
            measurement['resampling_method'] = 'nearest'
            measurement['src_varname'] = measurement['name']
        
        out = {'source_type': self.meta['name'],
               'output_type': product_name,
               'description': self.meta['description'],
               'location': ingest_location,
               'file_path_template': file_path_template,
               'storage': self.meta['storage'],
               'measurements': self.meta['measurements'],
               'global_attributes': global_attributes}
        
        with open(outname, 'w') as yml:
            yaml.dump(out, yml, default_flow_style=False)
    
    @property
    def measurements(self):
        """
        
        Returns
        -------
        dict of dict
            a dictionary with measurement names as keys
        """
        return dict([(x['name'], x) for x in self.meta['measurements']])
    
    def write(self, ymlfile):
        """
        write the product definition to a YML file
        
        Parameters
        ----------
        ymlfile: str
            the file to write to
        
        Returns
        -------
        
        """
        if os.path.isfile(ymlfile):
            raise RuntimeError('ingestion YML already exists: \n   {}'.format(ymlfile))
        
        self.__validate()
        with open(ymlfile, 'w') as yml:
            yaml.dump(self.meta, yml, default_flow_style=False)



# -*- coding: utf-8 -*-
###############################################################################
# pyroSAR configuration handling

# Copyright (c) 2018-2021, the pyroSAR Developers.

# This file is part of the pyroSAR Project. It is subject to the
# license terms in the LICENSE.txt file found in the top-level
# directory of this distribution and at
# https://github.com/johntruckenbrodt/pyroSAR/blob/master/LICENSE.txt.
# No part of the pyroSAR project, including this file, may be
# copied, modified, propagated, or distributed except according
# to the terms contained in the LICENSE.txt file.
###############################################################################
import os
import json

import configparser as ConfigParser

__LOCAL__ = ['sensor', 'projection', 'orbit', 'polarizations', 'acquisition_mode',
             'start', 'stop', 'product', 'spacing', 'samples', 'lines']

# a pattern to search for pyroSAR processing products and extract metadata attributes from the file name
product_pattern = r'(?:.*[/\\]|)' \
                  r'(?P<outname_base>' \
                  r'(?P<sensor>[A-Z0-9]{1,4})_+' \
                  r'(?P<acquisition_mode>[A-Z0-9]{1,4})_+' \
                  r'(?P<orbit>[AD])_' \
                  r'(?P<start>[0-9T]{15})' \
                  r'(?:_(?P<extensions>\w*?)|)' \
                  r')_*' \
                  r'(?:(?P<polarization>[HV]{2})_' \
                  r'(?P<proc_steps>[\w-]*)|)' \
                  r'(?P<filetype>(?:.tif|.nc|))$'


class Storage(dict):
    """
    Dict class with point access to store the lookups, pattern and URLs

    Attributes
    ----------
    STORAGE.LOOKUP : Storage
        All lookup table merged in a Storage class instance:
            * snap : SNAP process.
            * attributes : Attributes for different sensors.
    STORAGE.URL : dict (with point access)
        All URLs for DEMs, orbit files etc.:
            * dem : URL to download specific DEMs:
                * strm3
                * ace2
                * strm3_FTP
                * strm1HGT
                * ace
            * orbit : URL to download the orbit files:
                * ers1
                * ers2
                * s1_poe
                * s1_pres
                * doris
            * auxcal : URL to download the auxcal data:
                * s1
                * envisat
                * ers

    Note
    ----
    There may be additional attributes not listed above depending of the
    specific solver. Since this class is essentially a subclass of dict
    with attribute accessors, one can see which attributes are available
    using the `keys()` method.
    """
    
    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)
    
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__
    
    def __repr__(self):
        if self.keys():
            m = max(map(len, list(self.keys()))) + 1
            return '\n'.join([k.rjust(m) + ': ' + repr(v)
                              for k, v in sorted(self.items())])
        else:
            return self.__class__.__name__ + '()'
    
    def __dir__(self):
        return list(self.keys())


# ==============================================================================
# LOOKUP
# ==============================================================================
LOOKUP = Storage(attributes={'sensor': 'TEXT',
                             'orbit': 'TEXT',
                             'acquisition_mode': 'TEXT',
                             'start': 'TEXT',
                             'stop': 'TEXT',
                             'product': 'TEXT',
                             'samples': 'INTEGER',
                             'lines': 'INTEGER',
                             'outname_base': 'TEXT PRIMARY KEY',
                             'scene': 'TEXT',
                             'hh': 'INTEGER',
                             'vv': 'INTEGER',
                             'hv': 'INTEGER',
                             'vh': 'INTEGER'})

# ==============================================================================
# URL
# ==============================================================================
dem = Storage(ace2='https://step.esa.int/auxdata/dem/ACE2/5M/',
              ace='https://step.esa.int/auxdata/dem/ACE30/',
              srtm3_FTP='xftp.jrc.it',
              srtm3='https://srtm.csi.cgiar.org/wp-content/uploads/files/srtm_5x5/TIFF',
              srtm1Hgt='https://step.esa.int/auxdata/dem/SRTMGL1/', )

orbit = Storage(doris='https://step.esa.int/auxdata/orbits/Doris/vor',
                ers1='https://step.esa.int/auxdata/orbits/ers_precise_orb/ERS1',
                ers2='https://step.esa.int/auxdata/orbits/ers_precise_orb/ERS2',
                s1_poe='https://step.esa.int/auxdata/orbits/Sentinel-1/POEORB/',
                s1_res='https://step.esa.int/auxdata/orbits/Sentinel-1/RESORB/')

auxcal = Storage(s1='https://step.esa.int/auxdata/auxcal/S1/',
                 envisat='https://step.esa.int/auxdata/auxcal/ENVISAT/',
                 ers='https://step.esa.int/auxdata/auxcal/ERS/')

URL = Storage(dem=dem,
              orbit=orbit,
              auxcal=auxcal)

# ==============================================================================
# Merge
# ==============================================================================
STORAGE = Storage(URL=URL,
                  LOOKUP=LOOKUP)


class Singleton(type):
    """
    Define an Instance operation that lets clients access its unique instance.
    https://sourcemaking.com/design_patterns/singleton/python/1
    """
    
    def __init__(cls, name, bases, attrs, **kwargs):
        super().__init__(name, bases, attrs)
        cls._instance = None
    
    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__call__(*args, **kwargs)
        return cls._instance


class ConfigHandler(metaclass=Singleton):
    """
    ConfigHandler is a configuration handler for pyroSAR. It is intended to be called by a class's '__init__' and
    set or get the configuration parameters throughout an entire package.
    The primary goal with ConfigHandler is to load a single, consistent configuration environment to be passed 
    amongst ALL objects within a package.
        
    ConfigHandler is a SINGLETON, meaning once instantiated, THE SAME OBJECT
    will be returned to every class object calling it.

    Parameters
    ----------
    path : str or None
        A path where the .pyrosar directory will be created. If None (default) it will be created in the user home
        directory.
    config_fname : str
        Name of the config file. Default is 'config.ini'.
    
    Methods
    -------
    make_dir : Create a .pyrosar directory in home directory.
    create_config : Create a config.ini file in .pyrosar directory.
    open : Open the config.ini file.
    add_section : Create a new section in the configuration.
    set : Set an option in the configuration.
    remove_option : Remove an option in the configuration.

    Notes
    -----
    The syntax is the same as in ConfigParser. Here, keys are called options.

    """
    
    # Define __setter to control changeable keys (optional)
    # __setter = ["etc", "auxdata"]
    
    def __init__(self):
        path = os.path.join(os.path.expanduser('~'), '.pyrosar')
        
        self.__GLOBAL = {
            'path': path,
            'config_fname': 'config.ini',
            'config': os.path.join(path, 'config.ini'),
        }
        
        if not os.path.isfile(self.__GLOBAL['config']):
            self.__create_config()
        
        self.parser = ConfigParser.RawConfigParser(allow_no_value=True)
        self.parser.optionxform = str
        self.parser.read(self.__GLOBAL['config'])
    
    def __create_config(self):
        """
        Create a config.ini file in .pyrosar directory.

        Returns
        -------
        None
        """
        
        if not os.path.exists(self.__GLOBAL['path']):
            os.makedirs(self.__GLOBAL['path'])
        
        with open(self.__GLOBAL['config'], 'w'):
            pass
    
    def __str__(self):
        items = []
        for section in self.parser.sections():
            items.append('  Section: {0}\n'.format(section))
            
            for options in self.parser.options(section):
                items.append('    x {0} :: {1} :: {2}\n'
                             .format(options,
                                     self.parser.get(section, options),
                                     str(type(options))))
        out = 'Class    : Config\n' \
              'Path     : {0}\n' \
              'Sections : {1}\n' \
              'Contents : \n{2}' \
            .format(self.__GLOBAL['config'],
                    len(self.parser.sections()),
                    ''.join(items))
        
        return out
    
    def __getitem__(self, section):
        if not self.parser.has_section(section):
            raise AttributeError('Section {0} does not exist.'.format(str(section)))
        return dict(self.parser.items(section))
    
    @property
    def sections(self):
        return self.parser.sections()
    
    def keys(self, section):
        """
        Get all keys (options) of a section.

        Parameters
        ----------
        section : str
            Section name.

        Returns
        -------
        list : options (keys) of a section.

        """
        return self.parser.options(section)
    
    def open(self):
        """
        Open the config.ini file. This method will open the config.ini file in a external standard app (text editor).

        Returns
        -------
        os.startfile

        """
        
        os.startfile(self.__GLOBAL['config'])
    
    def add_section(self, section):
        """
        Create a new section in the configuration.

        Parameters
        ----------
        section : str
            Section name

        Returns
        -------
        None

        """
        if not self.parser.has_section(section):
            self.parser.add_section(section)
            self.write()
        else:
            raise RuntimeError('section already exists')
    
    @property
    def file(self):
        return self.__GLOBAL['config']
    
    def set(self, section, key, value, overwrite=False):
        """
        Set an option.

        Parameters
        ----------
        section : str
            Section name.
        key : str
            the attribute name
        value :
            the attribute value
        overwrite : bool
            If True and the defined key exists the value will be overwritten.

        Returns
        -------

        """
        if not self.parser.has_section(section):
            raise AttributeError('Section {0} does not exist.'.format(str(section)))
        
        if isinstance(value, list):
            value = json.dumps(value)
        
        if key in self.parser.options(section) and not overwrite:
            raise RuntimeError('Value already exists.')
        
        self.parser.set(section, key, value)
        self.write()
    
    def remove_option(self, section, key):
        """
        Remove an option and key.

        Parameters
        ----------
        section : str
            Section name.
        key : str
            Key value.
        
        Returns
        -------
        
        """
        if not self.parser.has_section(section):
            raise AttributeError('Section {0} does not exist.'.format(str(section)))
        
        if key not in self.parser.options(section):
            raise AttributeError('Key {0} does not exist.'.format(str(key)))
        
        self.parser.remove_option(section, key)
        self.write()
    
    def remove_section(self, section):
        """
        remove a section
        
        Parameters
        ----------
        section: str
            Section name.

        Returns
        -------

        """
        self.parser.remove_section(section)
        self.write()
    
    def write(self):
        with open(self.__GLOBAL['config'], 'w', encoding='utf8') as out:
            self.parser.write(out)



###############################################################################
# Utilities for simplification of lines used by pyroSAR for border noise removal

# Copyright (c) 2017-2020, the pyroSAR Developers.

# This file is part of the pyroSAR Project. It is subject to the
# license terms in the LICENSE.txt file found in the top-level
# directory of this distribution and at
# https://github.com/johntruckenbrodt/pyroSAR/blob/master/LICENSE.txt.
# No part of the pyroSAR project, including this file, may be
# copied, modified, propagated, or distributed except according
# to the terms contained in the LICENSE.txt file.
###############################################################################

from osgeo import ogr
import numpy as np
from spatialist.ancillary import rescale
from .polysimplify import VWSimplifier


import matplotlib
import matplotlib.pyplot as plt
from matplotlib.patches import Polygon
from matplotlib.collections import PatchCollection
matplotlib.rcParams['font.size'] = 12


def simplify(x, y, maxpoints=20):
    x = list(map(float, x))
    y = list(map(float, y))
    pts = np.array(list(zip(x, y)))
    simplifier = VWSimplifier(pts)
    sqd = []
    iter_range = range(2, maxpoints + 1)
    for i in iter_range:
        VWpts = simplifier.from_number(i)
        xn, yn = zip(*VWpts)
        out = np.sum((y - np.interp(x, xn, yn)) ** 2)
        sqd.append(out)
    # sqd /= max(sqd)
    if min(sqd) == max(sqd):
        VWpts = simplifier.from_number(2)
        return VWpts
    else:
        sqd = rescale(sqd)
        # plt.plot(sqd)
        # plt.show()
        # iter = (np.array(iter_range) - 2) / (maxpoints - 2.)
        # plt.plot(iter_range, sqd, label='residual')
        # plt.plot(iter_range, iter, color='r', label='iteration')
        # plt.plot(iter_range, iter + sqd, color='g', label='residual+iteration')
        # plt.legend(loc='upper center', shadow=True)
        # plt.show()
        # npoints = np.argmin(iter + sqd) + 2
        npoints = np.argmax(np.array(sqd) < 0.01) + 2
        VWpts = simplifier.from_number(npoints)
        return VWpts


def createPoly(xn, yn, xmax, ymax, plot=False):
    """
    create an OGR geometry from a sequence of indices
    
    Parameters
    ----------
    xn: numpy.ndarray
        the x indices of the points
    yn: numpy.ndarray
        the y  indices of the points
    xmax: int or float
        the maximum x index value
    ymax: int or float
        the maximum y index value

    Returns
    -------
    ogr.Geometry
    """
    ring = ogr.Geometry(ogr.wkbLinearRing)
    ring.AddPoint_2D(0, 0)
    for item in zip(xn, yn):
        item = list(map(int, item))
        if item != [0, 0] and item != [xmax, ymax]:
            ring.AddPoint_2D(item[0], item[1])
    ring.AddPoint_2D(xmax, ymax)
    ring.AddPoint_2D(xmax, 0)
    ring.CloseRings()
    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AddGeometry(ring)
    if plot:
        fig, ax = plt.subplots()
        pts = ring.GetPoints()
        arr = np.array(pts)
        polygon = Polygon(arr, True)
        p = PatchCollection([polygon], cmap=matplotlib.cm.jet, alpha=0.4)
        ax.add_collection(p)
        ax.autoscale_view()
        plt.scatter(arr[:, 0], arr[:, 1], s=10, color='red')
        plt.show()
    return poly


def reduce(seq, maxpoints=20, straighten=False, plot=False):
    """
    reduce the complexity of a line; the following steps are performed:
     - simplify the line using the Visvalingam-Whyatt method
     - iteratively add points on the original line back to the simplified line
       until the polygon spanned by the simplified line and (xmin, ymin) does not
       contain any further points of the original line; the polygon area is
       expected to only cover valid pixels of the image
     - optionally further straighten the result for smoother edges
    
    Parameters
    ----------
    seq: numpy.ndarray
        the 1D line sequence to be simplified
    maxpoints: int
        the maximum number points in the simplified sequence
    straighten: bool
        perform additional straightening on the simplified line?
    plot: bool
        plot the results?
    
    Returns
    -------
    numpy.ndarray
        the simplified line sequence
    """
    if min(seq) == max(seq):
        return np.array(seq)
    x = list(range(0, len(seq)))
    if plot:
        plt.plot(seq, label='ESA-corrected')
    # simplify the sequence using the Visvalingam-Whyatt algorithm
    VWpts = simplify(x, seq, maxpoints)
    xn, yn = [list(x) for x in zip(*VWpts)]
    if plot:
        plt.plot(xn, yn, linewidth=2, color='r', label='VW-simplified')
    simple = np.interp(x, xn, yn)
    # create a list of OGR points for the original border
    points = []
    for xi, yi in enumerate(seq):
        point = ogr.Geometry(ogr.wkbPoint)
        point.AddPoint(int(xi), int(yi))
        points.append(point)
    points = np.array(points)
    while True:
        # create a polygon containing all pixels inside the simplified border
        # i.e., containing the area considered valid
        poly = createPoly(xn, yn, seq.size, int(max(seq)))
        # create an OGR line from the simplified border points
        line = ogr.Geometry(ogr.wkbLineString)
        for xi, yi in zip(xn, yn):
            line.AddPoint(xi, yi)
        # compute the distance of each original point to the simplified line
        dists = np.array([line.Distance(point) for point in points])
        # check which points are inside of the polygon
        contain = np.array([point.Within(poly) for point in points])
        # remove points outside the polygon and stop if
        # no further points outside the polygon exist
        dists[~contain] = 0
        points = points[(dists > 0)]
        dists = dists[(dists > 0)]
        if len(dists) == 0:
            break
        # select the point with the largest distance to the simplified
        # line and add it to the list of simplified points
        # this reduces the size of the polygon an thus the area considered valid
        candidate = points[np.argmax(dists)]
        cp = candidate.GetPoint()
        index = np.argmin(np.array(xn) < cp[0])
        xn.insert(index, cp[0])
        yn.insert(index, cp[1])
    if plot:
        plt.plot(xn, yn, linewidth=2, color='limegreen', label='corrected')
    
    # further straighten the line segments
    # def straight(xn, yn, VWpts):
    #     indices = [i for i in range(0, len(xn)) if (xn[i], yn[i]) in VWpts]
    #     log.info(indices)
    #     for i, j in enumerate(indices):
    #         if i < (len(indices) - 1):
    #             if indices[i + 1] > j + 1:
    #                 dx = abs(xn[j] - xn[indices[i + 1]])
    #                 dy = abs(yn[j] - yn[indices[i + 1]])
    #                 if dx > dy:
    #                     seg_y = yn[j:indices[i + 1] + 1]
    #                     for k in range(j, indices[i + 1] + 1):
    #                         yn[k] = min(seg_y)
    #     return yn
    
    def straight(xn, yn, VWpts):
        indices = [i for i in range(0, len(xn)) if (xn[i], yn[i]) in VWpts]
        xn_new = []
        yn_new = []
        # make all line segments horizontal or vertical
        for index in range(len(indices) - 1):
            i = indices[index]
            j = indices[index + 1]
            ymin = min(yn[i:j + 1])
            xn_new.extend([xn[i], xn[j]])
            yn_new.extend([ymin, ymin])
        # shift horizontal lines down if the preceding horizontal line has a lower y value
        # but only if the shift is less than the tolerance
        tolerance = 15
        for i in range(len(xn_new) - 2):
            if yn_new[i] == yn_new[i + 1]:
                if yn_new[i] < yn_new[i + 2] and abs(yn_new[i] - yn_new[i + 2]) < tolerance:
                    yn_new[i + 2] = yn_new[i]
                    yn_new[i + 3] = yn_new[i]
                elif (yn_new[i] > yn_new[i + 2]) \
                        and (yn_new[i + 2] == yn_new[i + 3]) \
                        and abs(yn_new[i] - yn_new[i + 2]) < tolerance:
                    yn_new[i] = yn_new[i + 2]
                    yn_new[i + 1] = yn_new[i + 2]
        return xn_new, yn_new
    
    if straighten:
        xn, yn = straight(xn, yn, VWpts)
        if plot:
            plt.plot(xn, yn, linewidth=2, color='m', label='straightened')
    if plot:
        plt.legend()
        plt.xlabel('row')
        plt.ylabel('column')
        plt.show()
    return np.interp(x, xn, yn).astype(int)



from setuptools import setup, find_packages
import os

# Create .pyrosar in HOME - Directory
directory = os.path.join(os.path.expanduser('~'), '.pyrosar')
os.makedirs(directory, exist_ok=True)

directory = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='pyroSAR',
      packages=find_packages(),
      include_package_data=True,
      python_requires='>=3',
      setup_requires=['setuptools_scm'],
      use_scm_version=True,
      description='a framework for large-scale SAR satellite data processing',
      classifiers=[
          'License :: OSI Approved :: MIT License',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python :: 3'
      ],
      install_requires=['progressbar2',
                        'numpy',
                        'spatialist>=0.11.0',
                        'pyyaml',
                        'requests',
                        'psycopg2',
                        'SQLAlchemy>=1.4',
                        'SQLAlchemy-Utils>=0.37',
                        'GeoAlchemy2',
                        'lxml'],
      extras_require={
          'docs': ['sphinx', 'sphinxcontrib-bibtex',
                   'cairosvg', 'sphinxcontrib-svg2pdfconverter'],
      },
      url='https://github.com/johntruckenbrodt/pyroSAR.git',
      author='John Truckenbrodt',
      author_email='john.truckenbrodt@uni-jena.de',
      license='MIT',
      zip_safe=False,
      long_description=long_description,
      long_description_content_type='text/markdown')



###############################################################################
# tools for handling auxiliary data in software pyroSAR

# Copyright (c) 2019-2022, the pyroSAR Developers.

# This file is part of the pyroSAR Project. It is subject to the
# license terms in the LICENSE.txt file found in the top-level
# directory of this distribution and at
# https://github.com/johntruckenbrodt/pyroSAR/blob/master/LICENSE.txt.
# No part of the pyroSAR project, including this file, may be
# copied, modified, propagated, or distributed except according
# to the terms contained in the LICENSE.txt file.
###############################################################################
import io
import os
import re
import csv
import ssl
import ftplib
import requests
import zipfile as zf
from math import ceil, floor
from urllib.parse import urlparse
from lxml import etree as ET
from pyroSAR.examine import ExamineSnap
from spatialist.raster import Raster, Dtype
from spatialist.ancillary import dissolve, finder
from spatialist.auxil import gdalbuildvrt, crsConvert, gdalwarp
from spatialist.envi import HDRobject
from osgeo import gdal

import logging

log = logging.getLogger(__name__)


def dem_autoload(geometries, demType, vrt=None, buffer=None, username=None, password=None,
                 product='dem', nodata=None, dst_nodata=None, hide_nodata=False):
    """
    obtain all relevant DEM tiles for selected geometries

    Parameters
    ----------
    geometries: list[spatialist.vector.Vector]
        a list of :class:`spatialist.vector.Vector` geometries to obtain DEM data for;
        CRS must be WGS84 LatLon (EPSG 4326)
    demType: str
        the type of DEM to be used; current options:

        - 'AW3D30' (ALOS Global Digital Surface Model "ALOS World 3D - 30m")

          * info: https://www.eorc.jaxa.jp/ALOS/en/aw3d30/index.htm
          * url: ftp://ftp.eorc.jaxa.jp/pub/ALOS/ext1/AW3D30/release_v1804
          * height reference: EGM96

        - 'Copernicus 10m EEA DEM' (Copernicus 10 m DEM available over EEA-39 countries)

          * registration: https://spacedata.copernicus.eu/web/cscda/data-access/registration
          * url: ftps://cdsdata.copernicus.eu/DEM-datasets/COP-DEM_EEA-10-DGED/2021_1
          * height reference: EGM2008

        - 'Copernicus 30m Global DEM'
          
          * info: https://copernicus-dem-30m.s3.amazonaws.com/readme.html
          * url: https://copernicus-dem-30m.s3.eu-central-1.amazonaws.com/
          * height reference: EGM2008

        - 'Copernicus 30m Global DEM II'
        
          * registration: https://spacedata.copernicus.eu/web/cscda/data-access/registration
          * url: ftps://cdsdata.copernicus.eu/DEM-datasets/COP-DEM_GLO-30-DGED/2021_1
          * height reference: EGM2008
        
        - 'Copernicus 90m Global DEM'
     
          * info: https://copernicus-dem-90m.s3.amazonaws.com/readme.html
          * url: https://copernicus-dem-90m.s3.eu-central-1.amazonaws.com/
          * height reference: EGM2008
        
        - 'Copernicus 90m Global DEM II'
        
          * registration: https://spacedata.copernicus.eu/web/cscda/data-access/registration
          * url: ftps://cdsdata.copernicus.eu/DEM-datasets/COP-DEM_GLO-90-DGED/2021_1
          * height reference: EGM2008
        
        - 'GETASSE30'
        
          * info: https://seadas.gsfc.nasa.gov/help-8.1.0/desktop/GETASSE30ElevationModel.html
          * url: https://step.esa.int/auxdata/dem/GETASSE30
          * height reference: WGS84
        
        - 'SRTM 1Sec HGT'

          * url: https://step.esa.int/auxdata/dem/SRTMGL1
          * height reference: EGM96

        - 'SRTM 3Sec'

          * url: https://srtm.csi.cgiar.org/wp-content/uploads/files/srtm_5x5/TIFF
          * height reference: EGM96
        
        - 'TDX90m'
        
          * registration:  https://geoservice.dlr.de/web/dataguide/tdm90
          * url: ftpes://tandemx-90m.dlr.de
          * height reference: WGS84

    vrt: str or None
        an optional GDAL VRT file created from the obtained DEM tiles
    buffer: int, float, None
        a buffer in degrees to add around the individual geometries
    username: str or None
        (optional) the user name for services requiring registration
    password: str or None
        (optional) the password for the registration account
    product: str
        the sub-product to extract from the DEM product.
        The following options are available for the respective DEM types:
        
        - 'AW3D30'
        
          * 'dem': the actual Digital Elevation Model
          * 'msk': mask information for each pixel (Cloud/Snow Mask, Land water and
            low correlation mask, Sea mask, Information of elevation dataset used
            for the void-filling processing)
          * 'stk': number of DSM-scene files which were used to produce the 5 m resolution DSM
        
        - 'Copernicus 10m EEA DEM'
        
          * 'dem': the actual Digital Elevation Model
          * 'edm': editing mask
          * 'flm': filling mask
          * 'hem': height error mask
          * 'wbm': water body mask
        
        - 'Copernicus 30m Global DEM'
        
          * 'dem': the actual Digital Elevation Model
        
        - 'Copernicus 30m Global DEM II'
        
          * 'dem': the actual Digital Elevation Model
          * 'edm': editing mask
          * 'flm': filling mask
          * 'hem': height error mask
          * 'wbm': water body mask
        
        - 'Copernicus 90m Global DEM'
        
          * 'dem': the actual Digital Elevation Model
        
        - 'Copernicus 90m Global DEM II'
        
          * 'dem': the actual Digital Elevation Model
          * 'edm': editing mask
          * 'flm': filling mask
          * 'hem': height error mask
          * 'wbm': water body mask
        
        - 'GETASSE30'
        
          * 'dem': the actual Digital Elevation Model
        
        - 'SRTM 1Sec HGT'
        
          * 'dem': the actual Digital Elevation Model
        
        - 'SRTM 3Sec'
        
          * 'dem': the actual Digital Elevation Model
        
        - 'TDX90m'
        
          * 'dem': the actual Digital Elevation Model
          * 'am2': Amplitude Mosaic representing the minimum value
          * 'amp': Amplitude Mosaic representing the mean value
          * 'com': Consistency Mask
          * 'cov': Coverage Map
          * 'hem': Height Error Map
          * 'lsm': Layover and Shadow Mask, based on SRTM C-band and Globe DEM data
          * 'wam': Water Indication Mask
    
    nodata: int or float or None
        the no data value of the source files.
    dst_nodata: int or float or None
        the nodata value of the VRT file.
    hide_nodata: bool
        hide the VRT no data value?
    
    Returns
    -------
    list or None
        the names of the obtained files or None if a VRT file was defined
    
    Examples
    --------
    download all SRTM 1 arcsec DEMs overlapping with a Sentinel-1 scene and mosaic them to a single GeoTIFF file
    
    .. code-block:: python
        
        from pyroSAR import identify
        from pyroSAR.auxdata import dem_autoload
        from spatialist import gdalwarp
        
        # identify the SAR scene
        filename = 'S1A_IW_SLC__1SDV_20150330T170734_20150330T170801_005264_006A6C_DA69.zip'
        scene = identify(filename)
        
        # extract the bounding box as spatialist.Vector object
        bbox = scene.bbox()
        
        # download the tiles and virtually combine them in an in-memory
        # VRT file subsetted to the extent of the SAR scene plus a buffer of 0.01 degrees
        vrt = '/vsimem/srtm1.vrt'
        dem_autoload(geometries=[bbox], demType='SRTM 1Sec HGT',
                     vrt=vrt, buffer=0.01)
        
        # write the final GeoTIFF file
        outname = scene.outname_base() + 'srtm1.tif'
        gdalwarp(src=vrt, dst=outname, options={'format': 'GTiff'})
        
        # alternatively use function dem_create and warp the DEM to UTM
        # including conversion from geoid to ellipsoid heights
        from pyroSAR.auxdata import dem_create
        outname = scene.outname_base() + 'srtm1_ellp.tif'
        dem_create(src=vrt, dst=outname, t_srs=32632, tr=(30, 30),
                   geoid_convert=True, geoid='EGM96')
    """
    with DEMHandler(geometries) as handler:
        return handler.load(demType=demType,
                            username=username,
                            password=password,
                            vrt=vrt,
                            buffer=buffer,
                            product=product,
                            nodata=nodata,
                            dst_nodata=dst_nodata,
                            hide_nodata=hide_nodata)


def dem_create(src, dst, t_srs=None, tr=None, resampling_method='bilinear', threads=None,
               geoid_convert=False, geoid='EGM96', outputBounds=None, nodata=None,
               dtype=None, pbar=False):
    """
    create a new DEM GeoTIFF file and optionally convert heights from geoid to ellipsoid
    
    Parameters
    ----------
    src: str
        the input dataset, e.g. a VRT from function :func:`dem_autoload`
    dst: str
        the output dataset
    t_srs: None, int, str or osr.SpatialReference
        A target geographic reference system in WKT, EPSG, PROJ4 or OPENGIS format.
        See function :func:`spatialist.auxil.crsConvert()` for details.
        Default (None): use the crs of ``src``.
    tr: None or tuple
        the target resolution as (xres, yres)
    resampling_method: str
        the gdalwarp resampling method; See `here <https://gdal.org/programs/gdalwarp.html#cmdoption-gdalwarp-r>`_
        for options.
    threads: int, str or None
        the number of threads to use. Possible values:
        
         - Default `None`: use the value of `GDAL_NUM_THREADS` without modification. If `GDAL_NUM_THREADS` is None,
           multi-threading is still turned on and two threads are used, one for I/O and one for computation.
         - integer value: temporarily modify `GDAL_NUM_THREADS` and reset it once done.
           If 1, multithreading is turned off.
         - `ALL_CPUS`: special string to use all cores/CPUs of the computer; will also temporarily
           modify `GDAL_NUM_THREADS`.
    geoid_convert: bool
        convert geoid heights?
    geoid: str
        the geoid model to be corrected, only used if ``geoid_convert == True``; current options:
        
         - 'EGM96'
         - 'EGM2008'
    outputBounds: list or None
        output bounds as [xmin, ymin, xmax, ymax] in target SRS
    nodata: int or float or None
        the no data value of the source and destination files.
        Can be used if no source nodata value can be read or to override it.
    dtype: str or None
        override the data type of the written file; Default None: use same type as source data.
        Data type notations of GDAL (e.g. `Float32`) and numpy (e.g. `int8`) are supported.
    pbar: bool
        add a progressbar?
    
    Returns
    -------

    """
    
    with Raster(src) as ras:
        if nodata is None:
            nodata = ras.nodata
        epsg_in = ras.epsg
    
    if nodata is None:
        raise RuntimeError('the nodata value could not be read from the source file. Please explicitly define it.')
    
    if t_srs is None:
        epsg_out = epsg_in
    else:
        epsg_out = crsConvert(t_srs, 'epsg')
    
    threads_system = gdal.GetConfigOption('GDAL_NUM_THREADS')
    if threads is None:
        threads = threads_system
        try:
            threads = int(threads)
        except (ValueError, TypeError):
            pass
    if isinstance(threads, str):
        if threads != 'ALL_CPUS':
            raise ValueError("unsupported value for 'threads': '{}'".format(threads))
        else:
            multithread = True
            gdal.SetConfigOption('GDAL_NUM_THREADS', threads)
    elif isinstance(threads, int):
        if threads == 1:
            multithread = False
        elif threads > 1:
            multithread = True
            gdal.SetConfigOption('GDAL_NUM_THREADS', str(threads))
        else:
            raise ValueError("if 'threads' is of type int, it must be >= 1")
    elif threads is None:
        multithread = True
    else:
        raise TypeError("'threads' must be of type int, str or None. Is: {}".format(type(threads)))
    
    gdalwarp_args = {'format': 'GTiff', 'multithread': multithread,
                     'srcNodata': nodata, 'dstNodata': nodata,
                     'srcSRS': 'EPSG:{}'.format(epsg_in),
                     'dstSRS': 'EPSG:{}'.format(epsg_out),
                     'resampleAlg': resampling_method}
    
    if dtype is not None:
        gdalwarp_args['outputType'] = Dtype(dtype).gdalint
    
    if outputBounds is not None:
        gdalwarp_args['outputBounds'] = outputBounds
    
    if tr is not None:
        gdalwarp_args.update({'xRes': tr[0],
                              'yRes': tr[1],
                              'targetAlignedPixels': True})
    
    if geoid_convert:
        geoid_epsg = {'EGM96': 5773,
                      'EGM2008': 3855}
        if geoid in geoid_epsg.keys():
            epsg = geoid_epsg[geoid]
            gdalwarp_args['srcSRS'] += '+{}'.format(epsg)
            # the following line is a temporary workaround until compound EPSG codes can
            # directly be used for vertical CRS transformations
            # see https://github.com/OSGeo/gdal/pull/4639
            gdalwarp_args['srcSRS'] = crsConvert(gdalwarp_args['srcSRS'], 'proj4')
        else:
            raise RuntimeError('geoid model not yet supported')
        try:
            get_egm_lookup(geoid=geoid, software='PROJ')
        except OSError as e:
            errstr = str(e)
            addition = '\nplease refer to the following site for instructions ' \
                       'on how to use EGM GTX files (requires PROJ >= 5.0.0):\n' \
                       'https://gis.stackexchange.com/questions/258532/' \
                       'noaa-vdatum-gdal-variable-paths-for-linux-ubuntu'
            raise RuntimeError(errstr + addition)
    
    try:
        message = 'creating mosaic'
        crs = gdalwarp_args['dstSRS']
        if crs != 'EPSG:4326':
            message += ' and reprojecting to {}'.format(crs)
        message += ': {}'.format(dst)
        log.info(message)
        gdalwarp(src, dst, gdalwarp_args, pbar)
    except Exception:
        if os.path.isfile(dst):
            os.remove(dst)
        raise
    finally:
        gdal.SetConfigOption('GDAL_NUM_THREADS', threads_system)


class DEMHandler:
    """
    | An interface to obtain DEM data for selected geometries
    | The files are downloaded into the ESA SNAP auxdata directory structure
    
    Parameters
    ----------
    geometries: list of spatialist.vector.Vector
        a list of geometries
    """
    
    def __init__(self, geometries):
        if not isinstance(geometries, list):
            raise RuntimeError('geometries must be of type list')
        
        for geometry in geometries:
            if geometry.getProjection('epsg') != 4326:
                raise RuntimeError('input geometry CRS must be WGS84 LatLon (EPSG 4326)')
        
        self.geometries = geometries
        try:
            self.auxdatapath = ExamineSnap().auxdatapath
        except AttributeError:
            self.auxdatapath = os.path.join(os.path.expanduser('~'), '.snap', 'auxdata')
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        return
    
    @staticmethod
    def __applybuffer(extent, buffer):
        ext = dict(extent)
        if buffer is not None:
            ext['xmin'] -= buffer
            ext['xmax'] += buffer
            ext['ymin'] -= buffer
            ext['ymax'] += buffer
        return ext
    
    @staticmethod
    def __buildvrt(tiles, vrtfile, pattern, vsi, extent, nodata=None, dst_nodata=None, hide_nodata=False):
        if vsi is not None:
            locals = [vsi + x for x in dissolve([finder(x, [pattern]) for x in tiles])]
        else:
            locals = tiles
        with Raster(locals[0]) as ras:
            if nodata is None:
                nodata = ras.nodata
            xres, yres = ras.res
        opts = {'outputBounds': (extent['xmin'], extent['ymin'],
                                 extent['xmax'], extent['ymax']),
                'srcNodata': nodata, 'targetAlignedPixels': True,
                'xRes': xres, 'yRes': yres, 'hideNodata': hide_nodata
                }
        if dst_nodata is not None:
            opts['VRTNodata'] = dst_nodata
        gdalbuildvrt(src=locals, dst=vrtfile,
                     options=opts)
    
    def __commonextent(self, buffer=None):
        ext_new = {}
        for geo in self.geometries:
            if len(ext_new.keys()) == 0:
                ext_new = geo.extent
            else:
                for key in ['xmin', 'ymin']:
                    if geo.extent[key] > ext_new[key]:
                        ext_new[key] = geo.extent[key]
                for key in ['xmax', 'ymax']:
                    if geo.extent[key] < ext_new[key]:
                        ext_new[key] = geo.extent[key]
        ext_new = self.__applybuffer(ext_new, buffer)
        return ext_new
    
    @staticmethod
    def intrange(extent, step):
        """
        generate sequence of integer coordinates marking the tie points of the individual DEM tiles
        
        Parameters
        ----------
        extent: dict
            a dictionary with keys `xmin`, `xmax`, `ymin` and `ymax` with coordinates in EPSG:4326.
        step: int
            the sequence steps

        Returns
        -------
        tuple[range]
            the integer sequences as (latitude, longitude)
        """
        lat = range(floor(float(extent['ymin']) / step) * step,
                    ceil(float(extent['ymax']) / step) * step,
                    step)
        lon = range(floor(float(extent['xmin']) / step) * step,
                    ceil(float(extent['xmax']) / step) * step,
                    step)
        return lat, lon
    
    @staticmethod
    def __retrieve(url, filenames, outdir):
        files = list(set(filenames))
        os.makedirs(outdir, exist_ok=True)
        locals = []
        for file in files:
            remote = '{}/{}'.format(url, file)
            local = os.path.join(outdir, os.path.basename(file))
            if not os.path.isfile(local):
                log.info('{} <<-- {}'.format(local, remote))
                r = requests.get(remote)
                r.raise_for_status()
                with open(local, 'wb') as output:
                    output.write(r.content)
                r.close()
            if os.path.isfile(local):
                locals.append(local)
        return sorted(locals)
    
    def __retrieve_ftp(self, url, filenames, outdir, username, password, port=0):
        files = list(set(filenames))
        os.makedirs(outdir, exist_ok=True)
        
        parsed = urlparse(url)
        
        if parsed.scheme == 'ftpes':
            ftp = ftplib.FTP_TLS(parsed.netloc)
            try:
                ftp.login(username, password)  # login anonymously before securing control channel
            except ftplib.error_perm as e:
                raise RuntimeError(str(e))
            ftp.prot_p()  # switch to secure data connection.. IMPORTANT! Otherwise, only the user and password is encrypted and not all the file data.
        elif parsed.scheme == 'ftps':
            ftp = ImplicitFTP_TLS()
            ftp.connect(host=parsed.netloc, port=port)
            ftp.login(username, password)
        else:
            ftp = ftplib.FTP(parsed.netloc, timeout=100)
            ftp.login()
        if parsed.path != '':
            ftp.cwd(parsed.path)
        locals = []
        for product_remote in files:
            product_local = os.path.join(outdir, os.path.basename(product_remote))
            if not os.path.isfile(product_local):
                try:
                    targetlist = ftp.nlst(product_remote)
                except ftplib.error_temp:
                    continue
                address = '{}://{}/{}{}'.format(parsed.scheme, parsed.netloc,
                                                parsed.path + '/' if parsed.path != '' else '', product_remote)
                log.info('{} <<-- {}'.format(product_local, address))
                with open(product_local, 'wb') as myfile:
                    ftp.retrbinary('RETR {}'.format(product_remote), myfile.write)
            if os.path.isfile(product_local):
                locals.append(product_local)
        ftp.close()
        return sorted(locals)
    
    @property
    def config(self):
        return {
            'AW3D30': {'url': 'ftp://ftp.eorc.jaxa.jp/pub/ALOS/ext1/AW3D30/release_v1804',
                       'nodata': -9999,
                       'vsi': '/vsitar/',
                       'pattern': {'dem': '*DSM.tif',
                                   'msk': '*MSK.tif',
                                   'stk': '*STK.tif'}
                       },
            'Copernicus 10m EEA DEM': {'url': 'ftps://cdsdata.copernicus.eu/DEM-datasets/COP-DEM_EEA-10-DGED/2021_1',
                                       'nodata': -32767.0,
                                       'vsi': '/vsitar/',
                                       'port': 990,
                                       'pattern': {'dem': '*DEM.tif',
                                                   'edm': '*EDM.tif',
                                                   'flm': '*FLM.tif',
                                                   'hem': '*HEM.tif',
                                                   'wbm': '*WBM.tif'}
                                       },
            'Copernicus 30m Global DEM': {'url': 'https://copernicus-dem-30m.s3.eu-central-1.amazonaws.com',
                                          'nodata': None,
                                          'vsi': None,
                                          'pattern': {'dem': '*DSM*'}
                                          },
            'Copernicus 30m Global DEM II': {
                'url': 'ftps://cdsdata.copernicus.eu/DEM-datasets/COP-DEM_GLO-30-DGED/2021_1',
                'nodata': -32767.0,
                'vsi': '/vsitar/',
                'port': 990,
                'pattern': {'dem': '*DEM.tif',
                            'edm': '*EDM.tif',
                            'flm': '*FLM.tif',
                            'hem': '*HEM.tif',
                            'wbm': '*WBM.tif'}
            },
            'Copernicus 90m Global DEM': {'url': 'https://copernicus-dem-90m.s3.eu-central-1.amazonaws.com',
                                          'nodata': None,
                                          'vsi': None,
                                          'pattern': {'dem': '*DSM*'}
                                          },
            'Copernicus 90m Global DEM II': {
                'url': 'ftps://cdsdata.copernicus.eu/DEM-datasets/COP-DEM_GLO-90-DGED/2021_1',
                'nodata': -32767.0,
                'vsi': '/vsitar/',
                'port': 990,
                'pattern': {'dem': '*DEM.tif',
                            'edm': '*EDM.tif',
                            'flm': '*FLM.tif',
                            'hem': '*HEM.tif',
                            'wbm': '*WBM.tif'}
            },
            'GETASSE30': {'url': 'https://step.esa.int/auxdata/dem/GETASSE30',
                          'nodata': None,
                          'vsi': '/vsizip/',
                          'pattern': {'dem': '*.GETASSE30'}
                          },
            'SRTM 1Sec HGT': {'url': 'https://step.esa.int/auxdata/dem/SRTMGL1',
                              'nodata': -32768.0,
                              'vsi': '/vsizip/',
                              'pattern': {'dem': '*.hgt'}
                              },
            'SRTM 3Sec': {'url': 'https://srtm.csi.cgiar.org/wp-content/uploads/files/srtm_5x5/TIFF',
                          'nodata': -32768.0,
                          'vsi': '/vsizip/',
                          'pattern': {'dem': 'srtm*.tif'}
                          },
            'TDX90m': {'url': 'ftpes://tandemx-90m.dlr.de',
                       'nodata': -32767.0,
                       'vsi': '/vsizip/',
                       'pattern': {'dem': '*_DEM.tif',
                                   'am2': '*_AM2.tif',
                                   'amp': '*_AMP.tif',
                                   'com': '*_COM.tif',
                                   'cov': '*_COV.tif',
                                   'hem': '*_HEM.tif',
                                   'lsm': '*_LSM.tif',
                                   'wam': '*_WAM.tif'}
                       }
        }
    
    def load(self, demType, vrt=None, buffer=None, username=None, password=None,
             product='dem', nodata=None, dst_nodata=None, hide_nodata=False):
        """
        obtain DEM tiles for the given geometries
        
        Parameters
        ----------
        demType: str
            the type fo DEM to be used
        vrt: str or None
            an optional GDAL VRT file created from the obtained DEM tiles
        buffer: int, float, None
            a buffer in degrees to add around the individual geometries
        username: str or None
            the download account username
        password: str or None
            the download account password
        product: str
            the sub-product to extract from the DEM product
             - 'AW3D30'
             
              * 'dem': the actual Digital Elevation Model
              * 'msk': mask information for each pixel (Cloud/Snow Mask, Land water and
                low correlation mask, Sea mask, Information of elevation dataset used
                for the void-filling processing)
              * 'stk': number of DSM-scene files which were used to produce the 5m resolution DSM
             
             - 'Copernicus 10m EEA DEM'
             
              * 'dem': the actual Digital Elevation Model
              * 'edm': Editing Mask
              * 'flm': Filling Mask
              * 'hem': Height Error Mask
              * 'wbm': Water Body Mask
             
             - 'Copernicus 30m Global DEM'
             
              * 'dem': the actual Digital Elevation Model
             
             - 'Copernicus 30m Global DEM II'
             
              * 'dem': the actual Digital Elevation Model
              * 'edm': Editing Mask
              * 'flm': Filling Mask
              * 'hem': Height Error Mask
              * 'wbm': Water Body Mask
             
             - 'Copernicus 90m Global DEM'
             
              * 'dem': the actual Digital Elevation Model
             
             - 'Copernicus 90m Global DEM II'
             
              * 'dem': the actual Digital Elevation Model
              * 'edm': Editing Mask
              * 'flm': Filling Mask
              * 'hem': Height Error Mask
              * 'wbm': Water Body Mask
             
             - 'GETASSE30'
             
              * 'dem': the actual Digital Elevation Model
             
             - 'SRTM 1Sec HGT'
             
              * 'dem': the actual Digital Elevation Model
             
             - 'SRTM 3Sec'
             
              * 'dem': the actual Digital Elevation Model
             
             - 'TDX90m'
             
              * 'dem': the actual Digital Elevation Model
              * 'am2': Amplitude Mosaic representing the minimum value
              * 'amp': Amplitude Mosaic representing the mean value
              * 'com': Consistency Mask
              * 'cov': Coverage Map
              * 'hem': Height Error Map
              * 'lsm': Layover and Shadow Mask, based on SRTM C-band and Globe DEM data
              * 'wam': Water Indication Mask
        nodata: int or float or None
            the no data value to write in the VRT if it will be written.
            If `None`, the value of the source products is passed on.

        Returns
        -------
        list or None
            the names of the obtained files or None if a VRT file was defined
        """
        keys = self.config.keys()
        if demType not in keys:
            raise RuntimeError("demType '{}' is not supported\n  "
                               "possible options: '{}'"
                               .format(demType, "', '".join(keys)))
        
        products = self.config[demType]['pattern'].keys()
        if product not in products:
            raise RuntimeError("product '{0}' not available for demType '{1}'\n"
                               "  options: '{2}'".format(product, demType, "', '".join(products)))
        
        outdir = os.path.join(self.auxdatapath, 'dem', demType)
        
        remotes = []
        for geo in self.geometries:
            corners = self.__applybuffer(geo.extent, buffer)
            remotes.extend(self.remote_ids(corners, demType=demType,
                                           username=username, password=password))
        
        if len(remotes) == 0:
            raise RuntimeError('could not find DEM tiles for the area of interest')
        
        if demType in ['AW3D30', 'TDX90m', 'Copernicus 10m EEA DEM',
                       'Copernicus 30m Global DEM II', 'Copernicus 90m Global DEM II']:
            port = 0
            if 'port' in self.config[demType].keys():
                port = self.config[demType]['port']
            locals = self.__retrieve_ftp(self.config[demType]['url'], remotes, outdir,
                                         username=username, password=password, port=port)
        else:
            locals = self.__retrieve(self.config[demType]['url'], remotes, outdir)
        
        if demType == 'GETASSE30':
            for item in locals:
                getasse30_hdr(item)
        
        if nodata is None:
            if product == 'dem':
                nodata = self.config[demType]['nodata']
        
        if vrt is not None:
            self.__buildvrt(tiles=locals, vrtfile=vrt,
                            pattern=self.config[demType]['pattern'][product],
                            vsi=self.config[demType]['vsi'],
                            extent=self.__commonextent(buffer),
                            nodata=nodata, dst_nodata=dst_nodata,
                            hide_nodata=hide_nodata)
            return None
        return locals
    
    def remote_ids(self, extent, demType, username=None, password=None):
        """
        parse the names of the remote files overlapping with an area of interest

        Parameters
        ----------
        extent: dict
            the extent of the area of interest with keys xmin, xmax, ymin, ymax
        demType: str
            the type fo DEM to be used
        username: str or None
            the download account username
        password: str or None
            the download account password

        Returns
        -------
        str
            the sorted names of the remote files
        """
        
        def index(x=None, y=None, nx=3, ny=3, reverse=False):
            if reverse:
                pattern = '{c:0{n}d}{id}'
            else:
                pattern = '{id}{c:0{n}d}'
            if x is not None:
                xf = pattern.format(id='W' if x < 0 else 'E', c=abs(x), n=nx)
            else:
                xf = ''
            if y is not None:
                yf = pattern.format(id='S' if y < 0 else 'N', c=abs(y), n=ny)
            else:
                yf = ''
            return yf, xf
        
        def cop_dem_remotes(extent, arcsecs):
            lat, lon = self.intrange(extent, step=1)
            indices = [index(x, y, nx=3, ny=2)
                       for x in lon for y in lat]
            base = 'Copernicus_DSM_COG_{res}_{0}_00_{1}_00_DEM'
            skeleton = '{base}/{base}.tif'.format(base=base)
            candidates = [skeleton.format(res=arcsecs, *item) for item in indices]
            remotes = []
            for candidate in candidates:
                response = requests.get(self.config[demType]['url'],
                                        params={'prefix': candidate})
                xml = ET.fromstring(response.content)
                content = xml.findall('.//Contents', namespaces=xml.nsmap)
                if len(content) > 0:
                    remotes.append(candidate)
            return remotes
        
        if demType == 'SRTM 1Sec HGT':
            lat, lon = self.intrange(extent, step=1)
            remotes = ['{0}{1}.SRTMGL1.hgt.zip'.format(*index(x, y, nx=3, ny=2))
                       for x in lon for y in lat]
        
        elif demType == 'GETASSE30':
            lat, lon = self.intrange(extent, step=15)
            remotes = ['{0}{1}.zip'.format(*index(x, y, nx=3, ny=2, reverse=True))
                       for x in lon for y in lat]
        
        elif demType == 'TDX90m':
            lat, lon = self.intrange(extent, step=1)
            remotes = []
            for x in lon:
                xr = abs(x) // 10 * 10
                for y in lat:
                    yf, xf = index(x=x, y=y, nx=3, ny=2)
                    remotes.append('90mdem/DEM/{y}/{hem}{xr:03d}/TDM1_DEM__30_{y}{x}.zip'
                                   .format(x=xf, xr=xr, y=yf, hem=xf[0]))
        
        elif demType == 'AW3D30':
            remotes = []
            lat, lon = self.intrange(extent, step=1)
            for x in lon:
                for y in lat:
                    remotes.append(
                        '{0}{1}/{2}{3}.tar.gz'.format(*index(x // 5 * 5, y // 5 * 5),
                                                      *index(x, y)))
        
        elif demType == 'SRTM 3Sec':
            lat = range(int((60 - float(extent['ymin'])) // 5) + 1,
                        int((60 - float(extent['ymax'])) // 5) + 2)
            lon = range(int((float(extent['xmin']) + 180) // 5) + 1,
                        int((float(extent['xmax']) + 180) // 5) + 2)
            remotes = ['srtm_{:02d}_{:02d}.zip'.format(x, y) for x in lon for y in lat]
        
        elif demType in ['Copernicus 10m EEA DEM',
                         'Copernicus 30m Global DEM II',
                         'Copernicus 90m Global DEM II']:
            lat, lon = self.intrange(extent, step=1)
            indices = [''.join(index(x, y, nx=3, ny=2))
                       for x in lon for y in lat]
            
            ftp = ImplicitFTP_TLS()
            parsed = urlparse(self.config[demType]['url'])
            host = parsed.netloc
            path = parsed.path
            ftp.connect(host=host, port=self.config[demType]['port'])
            ftp.login(username, password)
            ftp.cwd(path)
            
            obj = io.BytesIO()
            ftp.retrbinary('RETR mapping.csv', obj.write)
            obj = obj.getvalue().decode('utf-8').splitlines()
            
            ids = []
            stream = csv.reader(obj, delimiter=';')
            for row in stream:
                if row[1] + row[2] in indices:
                    ids.append(row[0])
            
            remotes = []
            remotes_base = []
            
            def ftp_search(target, files):
                pattern = '|'.join(files)
                if target.endswith('/'):
                    content = ftp.nlst(target)
                    for item in content:
                        ftp_search(target + '/' + item, files)
                else:
                    if target.endswith('.tar') and re.search(pattern, target):
                        base = os.path.basename(target)
                        if base not in remotes_base:
                            remotes.append(target)
                            remotes_base.append(base)
            
            if len(ids) < len(indices):
                log.warning('the available DEM tiles do not fully cover the area of interest')
            
            if len(ids) > 0:
                ftp_search(path + '/', ids)
            ftp.quit()
        
        elif demType == 'Copernicus 30m Global DEM':
            remotes = cop_dem_remotes(extent, arcsecs=10)
        
        elif demType == 'Copernicus 90m Global DEM':
            remotes = cop_dem_remotes(extent, arcsecs=30)
        
        else:
            raise ValueError('unknown demType: {}'.format(demType))
        
        return sorted(remotes)


def getasse30_hdr(fname):
    """
    create an ENVI HDR file for zipped GETASSE30 DEM tiles
    
    Parameters
    ----------
    fname: str
        the name of the zipped tile

    Returns
    -------

    """
    basename = os.path.basename(fname)
    pattern = r'(?P<lat>[0-9]{2})' \
              '(?P<ns>[A-Z])' \
              '(?P<lon>[0-9]{3})' \
              '(?P<ew>[A-Z]).zip'
    match = re.search(pattern, basename).groupdict()
    
    lon = float(match['lon'])
    if match['ew'] == 'W':
        lon *= -1
    lat = float(match['lat'])
    if match['ns'] == 'S':
        lat *= -1
    posting = 30 / 3600  # 30 arc seconds
    pixels = 1800
    
    map_info = ['Geographic Lat/Lon', '1.0000', '1.0000',
                str(lon),
                str(lat + pixels * posting),
                str(posting),
                str(posting),
                'WGS-84', 'units=Degrees']
    
    with zf.ZipFile(fname, 'a') as zip:
        files = zip.namelist()
        hdr = basename.replace('.zip', '.hdr')
        if hdr not in files:
            with HDRobject() as obj:
                obj.samples = pixels
                obj.lines = pixels
                obj.byte_order = 1
                obj.data_type = 2
                obj.map_info = '{{{}}}'.format(','.join(map_info))
                obj.coordinate_system_string = crsConvert(4326, 'wkt')
                zip.writestr(hdr, str(obj))


def get_egm_lookup(geoid, software):
    """
    Download lookup tables for converting EGM geoid heights to WGS84 ellipsoid heights.
    
    Parameters
    ----------
    geoid: str
        the geoid model; current options:
        
        - SNAP: 'EGM96'
        - PROJ: 'EGM96', 'EGM2008'
    software: str
        the software for which to download the EGM lookup
        
        - SNAP: default directory: ``~/.snap/auxdata/dem/egm96``; URL:
        
          * https://step.esa.int/auxdata/dem/egm96/ww15mgh_b.zip
        - PROJ: requires the ``PROJ_LIB`` environment variable to be set as download directory; URLs:
        
          * https://cdn.proj.org/us_nga_egm96_15.tif
          * https://cdn.proj.org/us_nga_egm08_25.tif

    Returns
    -------

    """
    if software == 'SNAP':
        try:
            auxdatapath = ExamineSnap().auxdatapath
        except AttributeError:
            auxdatapath = os.path.join(os.path.expanduser('~'), '.snap', 'auxdata')
        local = os.path.join(auxdatapath, 'dem', 'egm96', 'ww15mgh_b.zip')
        os.makedirs(os.path.dirname(local), exist_ok=True)
        if not os.path.isfile(local):
            remote = 'https://step.esa.int/auxdata/dem/egm96/ww15mgh_b.zip'
            log.info('{} <<-- {}'.format(local, remote))
            r = requests.get(remote)
            r.raise_for_status()
            with open(local, 'wb') as out:
                out.write(r.content)
            r.close()
    
    elif software == 'PROJ':
        gtx_lookup = {'EGM96': 'us_nga_egm96_15.tif',
                      'EGM2008': 'us_nga_egm08_25.tif'}
        gtx_remote = 'https://cdn.proj.org/' + gtx_lookup[geoid]
        
        proj_lib = os.environ.get('PROJ_LIB')
        if proj_lib is not None:
            gtx_local = os.path.join(proj_lib, os.path.basename(gtx_remote))
            if not os.path.isfile(gtx_local):
                if not os.access(proj_lib, os.W_OK):
                    raise OSError("cannot write to 'PROJ_LIB' path: {}".format(proj_lib))
                log.info('{} <<-- {}'.format(gtx_local, gtx_remote))
                r = requests.get(gtx_remote)
                r.raise_for_status()
                with open(gtx_local, 'wb') as out:
                    out.write(r.content)
                r.close()
        else:
            raise RuntimeError("environment variable 'PROJ_LIB' not set")
    else:
        raise TypeError("software must be either 'SNAP' or 'PROJ'")


class ImplicitFTP_TLS(ftplib.FTP_TLS):
    """
    FTP_TLS subclass that automatically wraps sockets in SSL to support implicit FTPS.
    taken from https://stackoverflow.com/a/36049814
    """
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._sock = None
    
    @property
    def sock(self):
        """Return the socket."""
        return self._sock
    
    @sock.setter
    def sock(self, value):
        """When modifying the socket, ensure that it is ssl wrapped."""
        if value is not None and not isinstance(value, ssl.SSLSocket):
            value = self.context.wrap_socket(value)
        self._sock = value
